import argparse
import re
from binascii import unhexlify


class InvalidUUIDException(Exception):
    """Base class for Invalid HandleException -- The attribute handle given was not valid on this server."""

    def __init__(self, uuid):
        self.code = 0x01
        self.name = "Invalid UUID"
        self.description = "The supplied UUID (%s) is not a valid 16-bit or 128-bit UUID. "\
                           "(ex: 2a1b or 000AA000-0BB0-10C0-80A0-00805F9B34FB" % uuid

"""
Validators for various BLESuite input parameters supplied by the user. The use of these validators is mostly limited
to import/export and CLI features.
"""


class InvalidBDADDRException(Exception):

    def __init__(self, address):
        self.code = 0x02
        self.name = "Invalid BD_ADDR"
        self.description = "The supplied BD_ADDR (%s) is not valid. Use format 00:11:22:33:44:55" % address


class InvalidAddressTypeByName(Exception):

    def __init__(self, address_type):
        self.code = 0x02
        self.name = "Invalid address type name"
        self.description = "The supplied address type name (%s) is not valid. Options: ['public', 'random']" % address_type


class InvalidATTHandle(Exception):

    def __init__(self, handle):
        self.code = 0x03
        self.name = "Invalid ATT handle"
        self.description = "The supplied attribute handle (%s) is not valid. The integer value must be >= 1 and <= 0xffff" % handle


class InvalidATTSecurityMode(Exception):

    def __init__(self, mode, level):
        self.code = 0x04
        self.name = "Invalid ATT security mode"
        self.description = "The supplied attribute security mode (%s, %s) is not valid. Options: (Security Mode, Security Level)" \
                           "[(0,0) -- No Access, (1,1) -- Open, (1,2) -- Requires encryption and does not require" \
                           " authenticated pairing," \
                           "(1,3) -- Requires encryption and requires authenticated pairing, (1,4) -- Requires encryption" \
                           " with Secure Connections pairing, (2,1) -- Data signing and does not require authenticated" \
                           " pairing, (2,2) -- Data signing and requires authenticated pairing]" % (mode, level)


class InvalidATTProperty(Exception):

    def __init__(self, att_property):
        self.code = 0x05
        self.name = "Invalid ATT property"
        self.description = "The supplied attribute property (%s) is not valid. Options: ['read', 'write']" % att_property


class InvalidGATTProperty(Exception):

    def __init__(self, gatt_property):
        self.code = 0x06
        self.name = "Invalid GATT property"
        self.description = "The supplied GATT property (%s) is not valid. Options: ['broadcast', 'read', 'write', " \
                           "'notify', 'indicate', 'authenticated signed writes', 'extended properties']" % gatt_property


class InvalidSMLTK(Exception):

    def __init__(self, ltk):
        self.code = 0x07
        self.name = "Invalid SM LTK"
        self.description = "The supplied SM LTK (%s) is not valid. Expects hex string (ie 'AB7F2A...')" % ltk


class InvalidSMRandom(Exception):

    def __init__(self, rand):
        self.code = 0x08
        self.name = "Invalid SM Random"
        self.description = "The supplied SM Random (%s) is not valid. Expects hex string (ie 'AB7F2A...')" % rand


class InvalidSMIRK(Exception):

    def __init__(self, ltk):
        self.code = 0x09
        self.name = "Invalid SM IRK"
        self.description = "The supplied SM IRK (%s) is not valid. Expects hex string (ie 'AB7F2A...')" % ltk


class InvalidSMCSRK(Exception):

    def __init__(self, ltk):
        self.code = 0x0a
        self.name = "Invalid SM LTK"
        self.description = "The supplied SM CSRK (%s) is not valid. Expects hex string (ie 'AB7F2A...')" % ltk


def validate_bluetooth_address_cli(address):
    """
    Validates BT address string

    :param address: BT Address
    :type: str
    :return: address
    """
    if address is not None:
        match = re.search("^[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}$"
                          , address)
        if match is not None and match.group(0) is not None:
            return address

    raise argparse.ArgumentTypeError("%s is an invalid Bluetooth address (BD_ADDR)" % address)


def validate_bluetooth_address(address):
    """
    Validates BT address string

    :param address: BT Address
    :type: str
    :return: address
    """
    if address is not None:
        match = re.search("^[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}$"
                          , address)
        if match is not None and match.group(0) is not None:
            return address

    raise InvalidBDADDRException(address)


def validate_attribute_uuid(uuid):
    from whad.ble.profile.attribute import UUID
    if uuid is None:
        raise InvalidUUIDException(uuid)
    # When attribute UUID is read from JSON, it can be encoded as unicode, which will break the UUID class
    try:
        UUID(uuid)
    except InvalidUUIDException:
        raise InvalidUUIDException(uuid)
    return uuid


def validate_address_type_name(address_type_name):
    if address_type_name is None:
        return InvalidAddressTypeByName(address_type_name)
    address_type_name = address_type_name.lower()
    if address_type_name == "public" or address_type_name == "random":
        return address_type_name

    raise InvalidAddressTypeByName(address_type_name)


def validate_int_att_handle(int_handle):
    if int_handle is None or int_handle < 0x01 or int_handle > 0xffff:
        raise InvalidATTHandle
    return int_handle


def validate_att_security_mode(mode, level):
    supported_modes = [
        (0, 0), (1, 1), (1, 2), (1, 3), (1, 4), (2, 1), (2, 2)
    ]

    if (mode, level) not in supported_modes:
        raise InvalidATTSecurityMode(mode, level)
    return mode, level


def validate_att_property(att_property):
    att_property = att_property.lower()

    if att_property != "read" and att_property != "write":
        raise InvalidATTProperty(att_property)
    return att_property


def validate_gatt_property(gatt_property):
    gatt_property = gatt_property.lower()
    valid_properties = ['broadcast', 'read', 'write', 'notify', 'indicate',
                        'authenticated signed writes', 'extended properties',
                        'write without response']
    if gatt_property not in valid_properties:
        raise InvalidGATTProperty(gatt_property)
    return gatt_property


def validate_ltk(ltk):
    ltk = ltk.lower()
    try:
        unhexlify(ltk)
    except TypeError:
        raise InvalidSMLTK(ltk)
    return ltk


def validate_irk(irk):
    if irk is None:
        irk = "00" * 16
    irk = irk.lower()
    try:
        unhexlify(irk)
    except TypeError:
        raise InvalidSMIRK(irk)
    return irk


def validate_csrk(csrk):
    if csrk is None:
        csrk = "00" * 16
    csrk = csrk.lower()
    try:
        unhexlify(csrk)
    except TypeError:
        raise InvalidSMCSRK(csrk)
    return csrk


def validate_rand(rand):
    rand = rand.lower()
    try:
        unhexlify(rand)
    except TypeError:
        raise InvalidSMRandom(rand)
    return rand





from whad.esb.connector import ESB, PTX
from whad.esb.stack import ESBStack
from whad.exceptions import UnsupportedCapability, WhadDeviceDisconnected
from whad.esb.injecting import InjectionConfiguration
from whad.helpers import message_filter, is_message_type
from time import sleep, time


class Injector(PTX):
    """
    Enhanced ShockBurst Injector interface.
    """
    def __init__(self, device):
        super().__init__(device)
        self._synced = False
        self._configuration = InjectionConfiguration()
        self._injecting = False
        #self.start()

    @property
    def configuration(self) -> InjectionConfiguration:
        """Retrieve this injector configuration.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration: InjectionConfiguration):
        """Set the injector configuration.
        """
        self.stop()
        self.__configuration = new_configuration
        if self.__configuration.address is not None:
            self.address = self.__configuration.address
        if self.__configuration.channel is not None:
            self.channel = self.__configuration.channel

        self.start()

    def on_pdu(self, pdu):
        if self._injecting:
            return
        else:
            return super().on_pdu(pdu)

    def inject(self, packet):
        if hasattr(packet, "address") and packet.address != self.address:
            self.stop()
            self.address = packet.address
            self.start()
        if self.__configuration.synchronize:
            if not self._synced:
                self.synchronize()
                self._injecting = True
                while not self.send(packet, channel=self.channel, address=self.address):
                    self._injecting = False
                    self.synchronize()
                self._synced = True
                return True
            else:
                success = self.send(packet, channel=self.channel, address=self.address)
                return success
        else:
            success = self.send(packet, channel=self.__configuration.channel, address=self.address)



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
from prompt_toolkit import print_formatted_text, HTML

from whad.common.monitors.pcap import PcapWriterMonitor
from whad.cli.app import CommandLineApp, ApplicationError, run_app
from scapy.all import *
from whad.device.unix import  UnixSocketConnector
from whad.device import Bridge, ProtocolHub
from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady
from whad.cli.ui import error, warning, success, info, display_event, display_packet

import logging
import time
import sys

logger = logging.getLogger(__name__)
#logging.basicConfig(level=logging.DEBUG)

class WhadExtractUnixSocketConnector(UnixSocketConnector):
    pass

class WhadExtractApp(CommandLineApp):

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD filter tool',
            interface=True,
            commands=False,
            input=CommandLineApp.INPUT_WHAD,
            output=CommandLineApp.OUTPUT_STANDARD
        )

        self.add_argument(
            'extractor',
            help='extractor to evaluate',
            nargs='*'
        )

        self.add_argument(
            '-d',
            '-D',
            '--delimiter',
            dest='delimiter',
            default=" ",
            help='delimiter between extractor'
        )

    def build_extractors(self):
        extractor_template = "lambda p : %s"
        extractors = []
        for extractor in self.args.extractor:
            if "packet." in extractor:
                extractor.replace("packet.", "p.")
            elif "pkt." in extractor:
                extractor.replace("pkt.", "p.")
            extractors.append((extractor, eval(extractor_template % extractor)))

        return extractors

    def on_packet(self, pkt):
        extractors = self.build_extractors()
        output = []
        try:
            for extractor,processor in extractors:
                output.append(str(processor(pkt)))
            print(self.args.delimiter.join(output))
            sys.stdout.flush()
            return pkt
        except Exception as e:
            sys.stderr.write(f"Exception raised when evaluating {extractor}\n")
            sys.stderr.flush()
            return pkt

    def run(self):
        # Launch pre-run tasks
        self.pre_run()
        try:
            if self.is_piped_interface():
                interface = self.input_interface
            else:
                interface = self.interface

            if interface is not None:
                if not self.args.nocolor:
                    conf.color_theme = BrightTheme()

                parameters = self.args.__dict__

                connector = WhadExtractUnixSocketConnector(interface)
                for parameter_name, parameter_value in parameters.items():
                    connector.add_parameter(parameter_name, parameter_value)

                connector.domain = self.args.domain
                hub = ProtocolHub(2)
                connector.format = hub.get(self.args.domain).format

                #connector.translator = get_translator(self.args.domain)(connector.hub)
                #connector.format = connector.translator.format
                connector.on_packet = self.on_packet

                #interface.open()

                while interface.opened:
                    time.sleep(.1)

        except KeyboardInterrupt:
            # Launch post-run tasks
            self.post_run()


def wextract_main():
    app = WhadExtractApp()
    run_app(app)



from whad.hub.discovery import Capability
from whad.protocol.ble.ble_pb2 import Start, Stop, AdvMode, SetAdvData, \
    ScanMode, CentralMode, ConnectTo, SendPDU, PeripheralMode, Disconnect, \
    SetEncryption
from whad.ble.stack.constants import BT_MANUFACTURERS
from enum import IntEnum

class HCIInternalState(IntEnum):
    NONE = 0
    SCANNING = 1
    CENTRAL = 2
    PERIPHERAL = 3

LE_STATES = {
    0 : ("Non-connectable Advertising State", 0 , [Start, Stop, AdvMode, SetAdvData]),
    1 : ("Scannable Advertising State", 0, [Start, Stop, AdvMode, SetAdvData]),
    2 : ("Connectable Advertising State", Capability.SimulateRole, [Start, Stop, AdvMode, SetAdvData]),
    3 : ("High Duty Cycle Directed Advertising State", 0,  [Start, Stop, AdvMode, SetAdvData]),
    4 : ("Passive Scanning State", Capability.Scan, [Start, Stop, ScanMode]),
    5 : ("Active Scanning State", Capability.Scan,[Start, Stop, ScanMode]),
    6 : ("Initiating State and Connection State (Central Role)", Capability.SimulateRole, [Start, Stop, CentralMode, ConnectTo,Disconnect, SendPDU, SetEncryption]),
    7 : ("Connection State (Peripheral Role)", Capability.SimulateRole,  [Start, Stop, PeripheralMode, SendPDU, Disconnect, SetEncryption])
}

ADDRESS_MODIFICATION_VENDORS = [
    BT_MANUFACTURERS[0].encode("utf-8"),
    BT_MANUFACTURERS[10].encode("utf-8"),
    BT_MANUFACTURERS[13].encode("utf-8"),
    BT_MANUFACTURERS[15].encode("utf-8"),
    BT_MANUFACTURERS[18].encode("utf-8"),
    BT_MANUFACTURERS[48].encode("utf-8"),
    BT_MANUFACTURERS[57].encode("utf-8")
]



from whad.rf4ce.connector import RF4CE
from whad.exceptions import UnsupportedCapability
from whad.dot15d4.stack import Dot15d4Stack
from whad.dot15d4.stack.mac import MACManager
from whad.dot15d4.address import Dot15d4Address
from whad.rf4ce.stack.nwk import NWKManager

import logging

logger = logging.getLogger(__name__)

class Target(RF4CE):
    """
    RF4CE Target Node interface for compatible WHAD device.
    """
    def __init__(self, device, profiles=[]):
        RF4CE.__init__(self, device)

        # Check if device can act as target node

        if not self.can_be_coordinator():
            raise UnsupportedCapability("Target")

        # Stack initialization
        MACManager.add(NWKManager)
        self.__stack = Dot15d4Stack(self)

        # Channel initialization
        self.__channel = 15
        self.__channel_page = 0

        self.__stack.get_layer('mac').set_extended_address(Dot15d4Address("11:22:33:44:55:66:77:88").value)
        self.__stack.get_layer('mac').set_short_address(Dot15d4Address(0x0384).value)
        self.__stack.get_layer('mac').database.set("macPanId", 0x1234)
        #self.__stack.get_layer('mac').database.set("macAckTimeout", 0.5)

        for profile in profiles:
            self.__stack.get_layer('apl').add_profile(profile)

        self.enable_reception()


    @property
    def stack(self):
        return self.__stack

    def enable_reception(self):
        self.set_coordinator_mode(channel=self.__channel)

    def set_channel(self, channel=11):
        self.__channel = channel
        self.enable_reception()

    def perform_ed_scan(self, channel):
        if not self.can_perform_ed_scan():
            raise UnsupportedCapability("EnergyDetection")
        self.__channel = channel
        super().perform_ed_scan(channel)

    def set_channel_page(self, page=0):
        if page != 0:
            raise UnsupportedCapability("ChannelPageSelection")
        else:
            self.__channel_page = page

    def auto_discovery(self):
        self.__stack.get_layer('nwk').get_service('management').auto_discovery()

    def discovery_response(self, status, destination_address):
        self.__stack.get_layer('nwk').get_service('management').discovery_response(
            status,
            Dot15d4Address(destination_address).value
        )



    def get_channel(self):
        return self.__channel

    def get_channel_page(self):
        return self.__channel_page


    def send(self, packet):
        packet.reserved = 1
        packet.do_build()
        super().send(packet, channel=self.__channel)

    def on_pdu(self, pdu):
        if (
            hasattr(pdu,"metadata") and
            hasattr(pdu.metadata, "is_fcs_valid") and
            not pdu.metadata.is_fcs_valid
        ):
            return

        self.__stack.on_pdu(pdu)

    def on_ed_sample(self, timestamp, sample):
        self.__stack.on_ed_sample(timestamp, sample)



from whad.common.analyzer import TrafficAnalyzer
from scapy.layers.bluetooth4LE import BTLE_ADV_IND, BTLE_SCAN_RSP, BTLE_CONNECT_REQ
from whad.hub.ble.bdaddr import BDAddress
from scapy.layers.bluetooth import  ATT_Read_By_Group_Type_Request, ATT_Read_By_Group_Type_Response, \
                                    ATT_Read_By_Type_Request, ATT_Read_By_Type_Response, ATT_Error_Response, \
                                    ATT_Find_Information_Request, ATT_Find_Information_Response
from whad.ble.stack.gatt.message import GattReadByGroupTypeResponse, GattReadByTypeResponse, GattFindInfoResponse
from whad.ble.profile.attribute import UUID
from whad.ble.profile import GenericProfile
from whad.ble.profile.characteristic import Characteristic, CharacteristicDescriptor, ClientCharacteristicConfig, CharacteristicValue
from whad.ble.profile.service import PrimaryService, SecondaryService
from struct import unpack
from whad.ble.crypto import EncryptedSessionInitialization, LegacyPairingCracking, \
    LongTermKeyDistribution, IdentityResolvingKeyDistribution, ConnectionSignatureResolvingKeyDistribution
import json

class PeripheralInformation(TrafficAnalyzer):
    def reset(self):
        super().reset()
        self.adv_ind_list = []
        self.scan_rsp_list = []
        self.adv_data = None
        self.scan_rsp = None
        self.address = None

    @property
    def output(self):
        if self.scan_rsp is None:
            return {
                "adv_data" : self.adv_data,
                "bd_addr" : str(self.address),
                "addr_type" : self.address.type if self.address is not None else None
            }
        else:
            return {
                "adv_data" : self.adv_data,
                "scan_rsp" : self.scan_rsp,
                "bd_addr" : str(self.address),
                "addr_type" : self.address.type if self.address is not None else None
            }
    def process_packet(self, packet):
        if BTLE_ADV_IND in packet:
            self.trigger()
            self.adv_ind_list.append(packet)
        elif BTLE_SCAN_RSP in packet:
            self.trigger()
            self.scan_rsp_list.append(packet)
        elif BTLE_CONNECT_REQ in packet:
            self.trigger()
            self.address = BDAddress(packet.AdvA, random=(packet.TxAdd != 0))
            for adv_ind in self.adv_ind_list[::-1]:
                if adv_ind.AdvA == packet.AdvA:
                    self.adv_data = b"".join([bytes(i) for i in adv_ind.data])
                    break
            for scan_rsp in self.adv_ind_list[::-1]:
                if scan_rsp.AdvA == packet.AdvA:
                    self.scan_rsp = b"".join([bytes(i) for i in scan_rsp.data])
                    break
            if self.scan_rsp is not None and self.adv_data is not None:
                self.complete()

class ReadByGroupTypeDiscovery(TrafficAnalyzer):

    def reset(self):
        super().reset()
        self.uuid = None
        self.handle = None
        self.start_handle = None
        self.end_handle = None
        self.items = []

    @property
    def output(self):
        return {
            "uuid" : self.uuid,
            "handle" : self.handle,
            "start_handle" : self.start_handle,
            "end_handle" : self.end_handle,
            "items" : self.items
        }

    def process_packet(self, packet):
        if ATT_Read_By_Group_Type_Request in packet:

            if not self.triggered:
                self.start_handle = packet.start
                self.end_handle = packet.end
                self.uuid = packet.uuid
                self.trigger()
            else:
                if packet.end != self.end_handle or packet.uuid != self.uuid:
                    if len(self.items) > 0:
                        self.complete()

                    self.reset()
            self.mark_packet(packet)

        elif ATT_Read_By_Group_Type_Response in packet and self.triggered:
            response = GattReadByGroupTypeResponse.from_bytes(
                item_size = packet.length,
                data = packet.data
            )
            for item in response:
                self.items.append(item)
                self.handle = item.end + 1
                if item.end == 0xFFFF or item.end == self.end_handle:
                    self.complete()

            self.mark_packet(packet)

        elif ATT_Error_Response in packet and packet.request == 0x10 and packet.ecode == 0x0a and self.triggered:
            self.mark_packet(packet)
            self.complete()



class ReadByTypeDiscovery(TrafficAnalyzer):

    def reset(self):
        super().reset()
        self.uuid = None
        self.handle = None
        self.start_handle = None
        self.end_handle = None
        self.items = []


    @property
    def output(self):
        return {
            "uuid" : self.uuid,
            "handle" : self.handle,
            "start_handle" : self.start_handle,
            "end_handle" : self.end_handle,
            "items" : self.items
        }

    def process_packet(self, packet):
        if ATT_Read_By_Type_Request in packet:
            if not self.triggered:
                self.start_handle = packet.start
                self.end_handle = packet.end
                self.uuid = packet.uuid
                self.trigger()
            else:
                if packet.end != self.end_handle or packet.uuid != self.uuid:
                    self.reset()
                    self.start_handle = packet.start
                    self.end_handle = packet.end
                    self.uuid = packet.uuid
                    self.trigger()
            self.mark_packet(packet)

        elif ATT_Read_By_Type_Response in packet and self.triggered:
            response = GattReadByTypeResponse.from_bytes(
                item_size = packet[ATT_Read_By_Type_Response].len,
                data = bytes(packet[ATT_Read_By_Type_Response:])[1:]
            )
            for item in response:
                self.items.append(item)
            self.complete()
            self.mark_packet(packet)
        elif ATT_Error_Response in packet and packet.request == 0x08 and packet.ecode == 0x0a and self.triggered:
            self.complete()
            self.mark_packet(packet)
        else:
            pass

class FindInformationDiscovery(TrafficAnalyzer):

    def reset(self):
        super().reset()
        self.start_handle = None
        self.end_handle = None
        self.items = []


    @property
    def output(self):
        return {
            "start_handle" : self.start_handle,
            "end_handle" : self.end_handle,
            "items" : self.items
        }

    def process_packet(self, packet):
        if ATT_Find_Information_Request in packet:
            if not self.triggered:
                self.start_handle = packet.start
                self.end_handle = packet.end
                self.trigger()
            else:
                if packet.end != self.end_handle:
                    self.reset()
                    self.start_handle = packet.start
                    self.end_handle = packet.end
                    self.trigger()
            self.mark_packet(packet)
        elif ATT_Find_Information_Response in packet and self.triggered:
            response = GattFindInfoResponse.from_bytes(
                format = packet[ATT_Find_Information_Response].format,
                data = bytes(packet[ATT_Find_Information_Response:])[1:]
            )
            for item in response:
                self.items.append(item)
            self.complete()
            self.mark_packet(packet)
        elif ATT_Error_Response in packet and packet.request == 0x04 and self.triggered:
            self.complete()
            self.mark_packet(packet)

class PrimaryServicesDiscovery(ReadByGroupTypeDiscovery):
    def reset(self):
        super().reset()
        self.primary_services = []


    @property
    def output(self):
        return {
            "primary_services" : self.primary_services,
        }

    def complete(self):
        if self.uuid == 0x2800:
            self.primary_services = [
                PrimaryService(
                    uuid=UUID(item.value),
                    handle=item.handle,
                    end_handle=item.end
                )
                for item in self.items
            ]
            super().complete()
        else:
            self.reset()


class CharacteristicsDiscovery(ReadByTypeDiscovery):
    def reset(self):
        super().reset()
        self.characteristics = []

    @property
    def output(self):
        return {
            "characteristics" : self.characteristics,
        }

    def complete(self):
        if self.uuid == 0x2803:
            for item in self.items:
                charac_properties = item.value[0]
                charac_handle = item.handle
                charac_value_handle = unpack('<H', item.value[1:3])[0]
                charac_uuid = UUID(item.value[3:])
                charac = Characteristic(
                    uuid=charac_uuid,
                    properties=charac_properties
                )
                charac.handle = charac_handle
                charac.value_handle = charac_value_handle
                self.characteristics.append(charac)
            super().complete()
        else:
            self.reset()

class SecondaryServicesDiscovery(ReadByGroupTypeDiscovery):
    def reset(self):
        super().reset()
        self.secondary_services = []

    @property
    def output(self):
        return {
            "secondary_services" : self.secondary_services,
        }

    def complete(self):
        if self.uuid == 0x2801:
            self.secondary_services = [
                SecondaryService(
                    uuid=UUID(item.value),
                    handle=item.handle,
                    end_handle=item.end
                )
                for item in self.items
            ]
            super().complete()
        else:
            self.reset()

class ServicesDiscovery(TrafficAnalyzer):

    def reset(self):
        super().reset()
        self.primary_service_analyzer = PrimaryServicesDiscovery()
        self.secondary_service_analyzer = SecondaryServicesDiscovery()
        self.services = []

    @property
    def output(self):
        return {
            "services" : self.services,
        }

    def complete(self):
        super().complete()

    def process_packet(self, pkt):
        self.primary_service_analyzer.process_packet(pkt)
        self.secondary_service_analyzer.process_packet(pkt)

        if self.primary_service_analyzer.triggered:
            self.trigger()

        if self.primary_service_analyzer.completed:
            for packet in self.primary_service_analyzer.marked_packets:
                self.mark_packet(packet)
            self.services += self.primary_service_analyzer.primary_services
            self.complete()

        if self.secondary_service_analyzer.completed:
            for packet in self.secondary_service_analyzer.marked_packets:
                self.mark_packet(packet)
            self.services += self.secondary_service_analyzer.secondary_services
            self.complete()

class DescriptorsDiscovery(FindInformationDiscovery):
    def reset(self):
        super().reset()
        self.descriptors = []


    @property
    def output(self):
        return {
            "descriptors" : self.descriptors,
        }

    def complete(self):
        for item in self.items:
            if item.uuid  == UUID(0x2901) or item.uuid == UUID(0x2902):
                self.descriptors.append(item)

        if len(self.descriptors) > 0:
            super().complete()


class GATTServerDiscovery(TrafficAnalyzer):
    class InferredProfile(GenericProfile):
        def export_json(self):
            json_value = super().export_json()
            if hasattr(self, "devinfo"):
                profile = json.loads(json_value)
                profile["devinfo"] = self.devinfo
                for k in profile["devinfo"].keys():
                    if isinstance(profile["devinfo"][k], bytes):
                        profile["devinfo"][k] = profile["devinfo"][k].hex()
                json_value = json.dumps(profile)
            return json_value

    def reset(self):
        super().reset()
        self.primary_service_analyzer = PrimaryServicesDiscovery()
        self.secondary_service_analyzer = SecondaryServicesDiscovery()
        self.characteristics_analyzer = CharacteristicsDiscovery()
        self.descriptors_discovery = DescriptorsDiscovery()
        self.peripheral_information = PeripheralInformation()

        self.services_completed = False
        self.characteristics_completed = False

        self.profile = None
        self.services = []
        self.characteristics = []
        self.descriptors = []

    @property
    def output(self):

        return {
            "profile" : self.profile
        }

    def complete(self):
        super().complete()
        self.profile = self.InferredProfile()
        for service in self.services:
            self.profile.add_service(service)
        if self.peripheral_information.completed:
            self.profile.devinfo = self.peripheral_information.output

    def process_packet(self, pkt):
        self.primary_service_analyzer.process_packet(pkt)
        self.secondary_service_analyzer.process_packet(pkt)
        self.characteristics_analyzer.process_packet(pkt)
        self.descriptors_discovery.process_packet(pkt)
        self.peripheral_information.process_packet(pkt)


        if self.primary_service_analyzer.triggered:
            self.trigger()

        if self.primary_service_analyzer.completed:

            for packet in self.primary_service_analyzer.marked_packets:
                self.mark_packet(packet)

            self.services += self.primary_service_analyzer.primary_services
            self.services_completed = True
            self.primary_service_analyzer.reset()

        if self.secondary_service_analyzer.completed:

            for packet in self.secondary_service_analyzer.marked_packets:
                self.mark_packet(packet)

            self.services += self.secondary_service_analyzer.secondary_services
            self.secondary_service_analyzer.reset()

        if self.characteristics_analyzer.completed:

            for packet in self.characteristics_analyzer.marked_packets:
                self.mark_packet(packet)

            for characteristic in self.characteristics_analyzer.characteristics:
                for service in self.services:
                    if service.handle < characteristic.handle and characteristic.handle <= service.end_handle:
                        service.add_characteristic(characteristic)
            self.characteristics_completed = True
            self.characteristics_analyzer.reset()

        if self.descriptors_discovery.completed:

            for packet in self.descriptors_discovery.marked_packets:
                self.mark_packet(packet)

            for descriptor in self.descriptors_discovery.descriptors:
                self.descriptors += self.descriptors_discovery.descriptors
                selected_service = None
                for service in self.services:
                    if service.handle < descriptor.handle and descriptor.handle <= service.end_handle:
                        selected_service = service
                        break

                if selected_service is not None:
                    characteristics = list(selected_service.characteristics())
                else:
                    characteristics = []
                selected_characteristic = None
                for characteristic in characteristics:
                    if characteristic.handle <= descriptor.handle:
                        selected_characteristic = characteristic
                    else:
                        break
                if selected_characteristic is not None:
                    descriptor_obj = None
                    if descriptor.uuid  == UUID(0x2901):
                        descriptor_obj = CharacteristicDescriptor(
                                selected_characteristic,
                                handle=descriptor.handle,
                                uuid=descriptor.uuid
                        )
                    elif descriptor.uuid == UUID(0x2902):
                        descriptor_obj = ClientCharacteristicConfig(
                                selected_characteristic,
                                handle=descriptor.handle
                        )
                    if descriptor_obj is not None:
                        selected_characteristic.add_descriptor(descriptor_obj)

            self.descriptors_discovery.reset()


        if (self.triggered and
            self.characteristics_completed and
            not self.characteristics_analyzer.triggered and
            self.services_completed and
            not self.primary_service_analyzer.triggered and
            not self.descriptors_discovery.triggered
        ):
            handles = []
            for service in self.services:
                handles.append(service.handle)
                for characteristic in service.characteristics():
                    handles.append(characteristic.handle)
                    handles.append(characteristic.value_handle)

                    for descriptor in characteristic.descriptors():
                        handles.append(descriptor.handle)

                    if characteristic.can_notify() or characteristic.can_indicate():
                        if len(list(characteristic.descriptors())) == 0:
                            return

            if (
                    all([len(list(service.characteristics())) > 0 for service in self.services]) and
                    list(range(min(*handles), max(*handles)+1)) == handles
                ):
                self.complete()


analyzers = {
    "peripheral_information" : PeripheralInformation,
    "encrypted_session_initialization" : EncryptedSessionInitialization,
    "legacy_pairing_cracking" : LegacyPairingCracking,
    "ltk_distribution" : LongTermKeyDistribution,
    "irk_distribution" : IdentityResolvingKeyDistribution,
    "csrk_distribution" : ConnectionSignatureResolvingKeyDistribution,
    "profile_discovery" : GATTServerDiscovery
}



"""BLE profile command handler
"""
import logging

from time import time, sleep
from json import loads, dumps
from binascii import hexlify
from prompt_toolkit import print_formatted_text, HTML
from argparse import Namespace

from whad.ble.profile.characteristic import CharacteristicProperties
from whad.cli.app import command, CommandLineApp
from whad.ble import Central, Scanner
from whad.ble.exceptions import ConnectionLostException
from whad.hub.ble.bdaddr import BDAddress
from whad.ble.stack.att.exceptions import AttError
from whad.ble.stack.gatt.exceptions import GattTimeoutException
from whad.ble.cli.central.helpers import show_att_error

# Expected parameters that must be passed to our program
# to use an already established connection
EXPECTED_BLE_PARAMS = [
    "initiator_bdaddr",
    "initiator_addrtype",
    "target_bdaddr",
    "target_addrtype",
    "conn_handle",
]

def profile_discover(app: CommandLineApp, device):
    """Discover the GATT profile of a given device

    :param app: WHAD application instance
    :type app: :class:ẁhad.cli.app.CommandLineApp`
    :param device: Peripheral device to enumerate
    :type device: :class:`whad.ble.profile.PeripheralDevice`
    """
    try:
        device.discover()
    except AttError as atterr:
        show_att_error(app, atterr)
        return
    except GattTimeoutException:
        app.error("GATT Timeout while discovering services and characteristics. Aborted.")
        return

    # Show discovered services and characteristics
    for service in device.services():
        print_formatted_text(HTML((
            "<ansicyan><b>Service {name}</b></ansicyan> "
            " (handle {handle} to {end_handle})"
            )).format(
                name=service.name, handle=service.handle,
                end_handle=service.end_handle
            )
        )

        # Loop on service characteristics
        for charac in service.characteristics():
            properties = charac.properties
            rights = ''
            if properties & CharacteristicProperties.READ != 0:
                rights += '<ansigreen>R</ansigreen>'
            if properties & CharacteristicProperties.WRITE != 0:
                rights += '<ansired>W</ansired>'
            if properties & CharacteristicProperties.WRITE_WITHOUT_RESPONSE != 0:
                rights += '<ansimagenta>W</ansimagenta>'
            if properties & CharacteristicProperties.INDICATE != 0:
                rights += '<ansicyan>I</ansicyan>'
            if properties & CharacteristicProperties.NOTIFY != 0:
                rights += '<ansiblue>N</ansiblue>'

            # Print characteristic properties
            print_formatted_text(
                HTML("  <b>{name}</b> %s, handle {handle}, value handle: {value_handle}" % rights)
                    .format(
                        name=charac.name,
                        rights=rights,
                        handle=charac.handle,
                        value_handle=charac.value_handle
                    )
            )

            # Display descriptors
            for desc in charac.descriptors():
                print_formatted_text(
                    HTML("    <ansiblue>Descriptor type {name}</ansiblue>, handle: {handle}")
                        .format(name=desc.name, handle=desc.handle)
                )

        print('')

def profile_export(app, command_args: list, device, device_metadata: dict):
    """Export a discovered GATT profile into a JSON file.
    """
    # Load GATT profile JSON data
    json_data = device.export_json()
    profile = loads(json_data)

    # Add specific device info (for emulating)
    profile['devinfo'] = device_metadata
    json_data = dumps(profile)
    try:
        print(f"Writing profile JSON data to {command_args[0]} ...")
        open(command_args[0], 'w', encoding='utf-8').write(json_data)
    except IOError:
        app.error("An error occured when writing to %s", command_args[0])

@command('profile')
def profile_handler(app, command_args):
    """discover services and characteristics
    
    <ansicyan><b>profile</b> <i>[JSON_PROFILE]</i></ansicyan>

    This command connects to a target device, discover its services and characteristics,
    and optionnaly exports the device's profile into a <i>JSON_PROFILE</i> file. This JSON
    file may be used later for emulating a similar device with the exact same properties
    and profile, see <ansicyan>emulate</ansicyan> command.
    """
    # We need to have an interface specified
    if app.interface is not None and app.args.bdaddr is not None:

        # Switch to Scanner mode, search our device
        scanner = Scanner(app.interface)
        scanner.start()

        start_time = time()
        for device in scanner.discover_devices():
            if device.address.lower() == app.args.bdaddr.lower():
                if device.got_scan_rsp or (time() - start_time) >= 10.0:
                    break

        # Generate device advertising information
        device_metadata = {
            'adv_data': hexlify(device.adv_records.to_bytes()).decode('utf-8'),
            'scan_rsp': hexlify(device.scan_rsp_records.to_bytes()).decode('utf-8') if device.scan_rsp_records is not None else None,
            'bd_addr': str(device.address),
            'addr_type': device.address_type
        }
        scanner.stop()

        # Switch to central mode
        central = Central(app.interface)
        central.start()

        # Connect to target device
        device = central.connect(app.args.bdaddr, random=app.args.random)
        if device is None:
            app.error("Cannot connect to %s, device does not respond.", app.args.bdaddr)
        else:
            try:
                # Perform profile discovery
                profile_discover(app, device)

                # Export profile to JSON file if required
                if len(command_args) >= 1:
                    # Load GATT profile JSON data
                    json_data = device.export_json()
                    profile = loads(json_data)

                    # Add specific device info (for emulating)
                    profile['devinfo'] = device_metadata
                    json_data = dumps(profile)
                    try:
                        print('Writing profile JSON data to %s ...' % command_args[0])
                        open(command_args[0], 'w', encoding='utf-8').write(json_data)
                    except IOError:
                        app.error("An error occured when writing to %s", command_args[0])

                # Disconnect
                device.disconnect()
            except ConnectionLostException:
                app.error("BLE device disconnected during discovery.")

        # Terminate central
        central.stop()

    # Piped interface
    elif app.args.bdaddr is None and app.is_piped_interface():

        # Make sure we have all the required parameters
        for param in EXPECTED_BLE_PARAMS:
            if not hasattr(app.args, param):
                app.error("Source interface does not provide a BLE connection")

        initiator = BDAddress(str(app.args.initiator_bdaddr),
                              addr_type=int(app.args.initiator_addrtype))
        advertiser = BDAddress(str(app.args.initiator_bdaddr),
                               addr_type=int(app.args.initiator_addrtype))
        existing_connection = Namespace(
            initiator=initiator.value,
            init_addr_type=int(app.args.initiator_addrtype),
            advertiser=advertiser.value,
            adv_addr_type=int(app.args.target_addrtype),
            conn_handle=int(app.args.conn_handle)
        )

        central = Central(app.input_interface, existing_connection)
        device = central.peripheral()

        try:
            # Read GATT characteristic
            profile_discover(app, device)

            # Export profile to JSON file if required
            if len(command_args) >= 1:
                app.warning("Cannot create profile file when chained.")

        except ConnectionLostException:
            app.error("BLE device disconnected during discovery.")
    else:
        app.error("You need to specify a target device with option --bdaddr.")


import sys
import traceback
from threading import Thread, Lock
from queue import Queue, Empty
from binascii import hexlify
from time import time, sleep

# Whad imports
from whad.exceptions import RequiredImplementation, UnsupportedDomain, \
    WhadDeviceNotReady, WhadDeviceNotFound, WhadDeviceDisconnected, WhadDeviceTimeout, WhadDeviceError
from whad.protocol.generic_pb2 import ResultCode
from whad.protocol.whad_pb2 import Message
from whad.protocol.device_pb2 import Capability, DeviceDomainInfoResp, DeviceType, DeviceResetQuery
from whad.helpers import message_filter, asciiz
from whad.hub import ProtocolHub
from whad.hub.message import AbstractPacket, AbstractEvent
from whad.hub.generic.cmdresult import CommandResult, Success
from whad.hub.discovery import InfoQueryResp, DomainInfoQueryResp, DeviceReady

# Logging
import logging

# Remove scapy deprecation warnings
import warnings
from cryptography.utils import CryptographyDeprecationWarning
warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)

logger = logging.getLogger(__name__)

class WhadDeviceInfo(object):
    """This class caches a device information related to its firmware, type, and supported domains and capabilities.

    :param DeviceInfoResp info_resp:  Whad message containing the device basic information.
    """
    def __init__(self, info_resp):
        # Store device information
        self.__whad_version = info_resp.proto_min_ver
        self.__max_speed = info_resp.max_speed
        self.__fw_author = info_resp.fw_author
        self.__fw_url = info_resp.fw_url
        self.__fw_ver_maj = info_resp.fw_version_major
        self.__fw_ver_min = info_resp.fw_version_minor
        self.__fw_ver_rev = info_resp.fw_version_rev
        self.__device_type = info_resp.type
        self.__device_id = asciiz(info_resp.device_id)

        # Parse domains and capabilities
        self.__domains = {}
        self.__commands = {}
        for domain in info_resp.capabilities:
            self.__domains[domain & 0xFF000000] = domain & 0x00FFFFFF
            self.__commands[domain & 0xFF000000] = 0

    def add_supported_commands(self, domain, commands):
        """Adds supported command for a given domain to the device information.

        :param domain: target domain
        :param commands: bitmask representing the supported commands for the given domain
        """
        if domain in self.__domains:
            self.__commands[domain] = commands


    def has_domain(self, domain):
        """Determine if a domain is supported by this device.

        :param Domain domain: Domain to check.
        :returns: True if supported, False otherwise.
        """
        return domain in self.__domains


    def has_domain_cap(self, domain, capability):
        """Check if device supports a specific capability for a given domain.

        :param Domain domain: target domain
        :param Capability capability: capability to check
        """
        if domain in self.__domains:
            return (self.__domains[domain] & (1 << capability) > 0)
        return False


    def get_domain_capabilities(self, domain):
        """Get device domain capabilities.

        :param Domain domain: target domain
        :returns: Domain capabilities
        :rtype: Bitmask of capabilities
        """
        if domain in self.__domains:
            return self.__domains[domain]
        return None


    def get_domain_commands(self, domain):
        """Get supported commands for a specific domain

        :param Domain domain: Target domain
        :returns: Bitmask of supported commands
        """
        if domain in self.__commands:
            return self.__commands[domain]
        return None


    ##### Getters #####

    @property
    def version_str(self):
        """Returns the device firmware version string.

        :returns: Device firmware version string
        """
        return '%d.%d.%d' % (
            self.__fw_ver_maj,
            self.__fw_ver_min,
            self.__fw_ver_rev
        )


    @property
    def whad_version(self):
        """Returns the device supported whad version.
        """
        return self.__whad_version

    @property
    def fw_author(self):
        return self.__fw_author.decode('utf-8')

    @property
    def fw_url(self):
        return self.__fw_url.decode('utf-8')

    @property
    def max_speed(self):
        return self.__max_speed

    @property
    def device_type(self):
        """Returns the device type.
        """
        return self.__device_type

    @property
    def device_id(self):
        """Returns the device id.
        """
        return self.__device_id

    @property
    def domains(self):
        """Return the list of supported domains.
        """
        return self.__domains.keys()

class WhadDeviceConnector(object):
    """
    Device connector.

    A connector creates a link between a device and a protocol controller.
    """

    def __init__(self, device=None):
        """
        Constructor.

        Link the device with this connector, and this connector with the
        provided device.

        :param WhadDevice device: Device to be used with this connector.
        """
        self.__device = None
        self.set_device(device)
        if self.__device is not None:
            self.__device.set_connector(self)

        # Packet callbacks
        self.__callbacks_lock = Lock()
        self.__reception_callbacks = {}
        self.__transmission_callbacks = {}
        self.__error_callbacks = []

        # Connector lock mode
        self.__locked = False
        self.__locked_pdus = Queue()

        # Synchronous mode (not enabled by default)
        self.__synchronous = False
        self.__pending_pdus = Queue()


    def attach_error_callback(self, callback, context=None):
        '''Attach an error callback to this connector.

        :param callback: function handling errors.
        :param context:  context object to pass to the error handling function.
        :returns: Boolean indicating if the callback has been successfully attached.
        '''
        # Enter critical section
        self.__callbacks_lock.acquire()
        self.__error_callbacks.append(
            (callback, context)
        )

        # Leave critical section
        self.__callbacks_lock.release()

    def on_error(self, error):
        '''Triggers a call to the device connector error handling registered callback(s).
        '''
        if len(self.__error_callbacks) > 0:
            # Duplicate error callbacks list
            self.__callbacks_lock.acquire()
            callbacks = list(self.__error_callbacks)
            self.__callbacks_lock.release()

            # Call each error callback
            for callback, context in callbacks:
                callback(error, context=context)


    def attach_callback(self, callback, on_reception=True, on_transmission=True, filter=lambda pkt:True):
        """
        Attach a new packet callback to current connector.

        :param callback: Processing function.
        :param on_reception: Boolean indicating if the callback monitors reception.
        :param on_transmission: Boolean indicating if the callback monitors transmission.
        :param filter: Lambda function filtering packets matching the callback.
        :returns: Boolean indicating if the callback has been successfully attached.
        """
        # Enter critical section
        self.__callbacks_lock.acquire()

        callbacks_dicts = (
            ([self.__reception_callbacks] if on_reception else []) +
            ([self.__transmission_callbacks] if on_transmission else [])
        )
        for callback_dict in callbacks_dicts:
            callback_dict[callback] = filter

        # Leave critical section
        self.__callbacks_lock.release()

        return len(callbacks_dicts) > 0

    def detach_callback(self, callback, on_reception=True, on_transmission=True):
        """
        Detach an existing packet callback from current connector.

        :param callback: Processing function.
        :param on_reception: Boolean indicating if the callback was monitoring reception.
        :param on_transmission: Boolean indicating if the callback was monitoring transmission.
        :returns: Boolean indicating if the callback has been successfully detached.
        """
        # Enter critical section
        self.__callbacks_lock.acquire()

        removed = False
        callbacks_dicts = (
            ([self.__reception_callbacks] if on_reception else []) +
            ([self.__transmission_callbacks] if on_transmission else [])
        )
        for callback_dict in callbacks_dicts:
            if callback in callback_dict:
                del callback_dict[callback]
                removed = True

        # Leave critical section
        self.__callbacks_lock.release()

        return removed

    def reset_callbacks(self, reception = True, transmission = True):
        """
        Detach any packet callback attached to the current connector.

        :param on_reception: Boolean indicating if the callbacks monitoring reception are detached.
        :param on_transmission: Boolean indicating if the callbacks monitoring transmission are detached.
        :returns: Boolean indicating if callbacks have been successfully detached.
        """

        # Enter critical section
        self.__callbacks_lock.acquire()

        callbacks_dicts = (
            ([self.__reception_callbacks] if reception else []) +
            ([self.__transmission_callbacks] if transmission else [])
        )
        for callback_dict in callbacks_dicts:
            callback_dict = {}

        # Leave critical section
        self.__callbacks_lock.release()

        return len(callbacks_dicts) > 0


    def monitor_packet_tx(self, packet):
        """
        Signals the transmission of a packet and triggers execution of matching transmission callbacks.

        :param packet: scapy packet being transmitted from whad-client.
        """
        # Enter critical section
        self.__callbacks_lock.acquire()

        for callback,packet_filter in self.__transmission_callbacks.items():
            if packet_filter(packet):
                callback(packet)

        # Leave critical section
        self.__callbacks_lock.release()


    def monitor_packet_rx(self, packet):
        """
        Signals the reception of a packet and triggers execution of matching reception callbacks.

        :param packet: scapy packet being received by whad-client.
        """
        # Enter critical section
        self.__callbacks_lock.acquire()

        for callback,packet_filter in self.__reception_callbacks.items():
            if packet_filter(packet):
                callback(packet)

        # Leave critical section
        self.__callbacks_lock.release()


    def set_device(self, device=None):
        """
        Set device linked to this connector.

        :param WhadDevice device: Device to be used with this connector.
        """
        if device is not None:
            self.__device = device

    @property
    def device(self):
        return self.__device

    @property
    def hub(self):
        return self.__device.hub

    def enable_synchronous(self, enabled : bool):
        """Enable or disable synchronous mode

        Synchronous mode is a mode in which the connector expects sone third-party code to
        retrieve the received packets instead of forwarding them to the `on_packet()` callback.
        It is then possible to wait for some packet to be received and avoid the automatic
        behavior triggered by a call to `on_packet()`.

        :param enabled: If set to `True`, enable synchronous mode. Otherwise disable it.
        :type enabled: bool
        """
        # Clear pending packets if we are disabling this feature.
        if not enabled:
            self.__pending_pdus.queue.clear()

        # Update state
        self.__synchronous = enabled

    def is_synchronous(self):
        """Determine if the conncetor is in synchronous mode.

        :return: `True` if synchronous mode is enabled, `False` otherwise.
        """
        return self.__synchronous

    def add_pending_packet(self, pdu):
        """Add a pending protocol data unit (PDU) if in synchronous mode.

        :param pdu: Pending PDU to add to our queue of pending PDUs
        :type pdu: scapy.packet.Packet
        """
        if self.__synchronous:
            self.__pending_pdus.put(pdu)

    def wait_packet(self, timeout:float = None):
        '''Wait for a packet when in synchronous mode.

        :param timeout: If specified, defines a timeout when querying the PDU queue
        :type timeout: float, optional
        :return: Received packet if any, None otherwise
        :rtype: scapy.packet.Packet
        '''
        if self.__synchronous:
            try:
                return self.__pending_pdus.get(block=True, timeout=timeout)
            except Empty as no_pdu:
                return None
        else:
            return None

    def lock(self):
        """Lock connector. A locked connector will not dispatch packets/pdus like
        in synchronous mode and will keep them in a waiting queue, but will dispatch
        them all at once when unlocked.
        """
        self.__locked = True

        # Clear pending PDUs queue
        with self.__locked_pdus.mutex:
            self.__locked_pdus.queue.clear()

    def unlock(self, dispatch_callback=None):
        """Unlock connector and dispatch pending PDUs.

        :param  dispatch_callback: PDU dispatch callback that overrides the internal dispatch routine
        :type   dispatch_callback: callable
        """
        # Mark connector as unlocked
        self.__locked = False

        # Dispatch PDUs
        try:
            while True:
                # Retrieve PDU
                pdu = self.__locked_pdus.get(block=False, timeout=0.2)

                if dispatch_callback is None:
                    # If connector is in synchronous mode, move PDUs to our pending queue
                    if self.__synchronous:
                        self.add_pending_packet(pdu)
                    else:
                        # Else forward PDU to our standard PDU processing method
                        self.on_packet(pdu)
                else:
                    # Call the provided dispatch callback
                    dispatch_callback(pdu)
        except Empty:
            # Processing done, continue.
            pass

    def is_locked(self) -> bool:
        """Determine if the connector is locked.

        :return: `True` if lock mode is enabled, `False` otherwise.
        """
        return self.__locked
    
    def add_locked_pdu(self, pdu):
        """Add a pending Protocol Data Unit (PDU) to our locked pdus queue.

        :param  Packet pdu:  Packet to add to locked packets queue
        :type   pdu: scapy.packet.Packet
        """
        self.__locked_pdus.put(pdu)

    # Device interaction
    def send_message(self, message, filter=None):
        """Sends a message to the underlying device without waiting for an answer.

        :param Message message: WHAD message to send to the device.
        :param filter: optional filter function for incoming message queue.
        """
        try:
            logger.debug('sending WHAD message to device: %s' % message)
            return self.__device.send_message(message, filter)
        except WhadDeviceError as device_error:
            logger.debug('an error occured while communicating with the WHAD device !')
            self.on_error(device_error)

    def send_command(self, message, filter=None):
        """Sends a command message to the underlying device and waits for an answer.

        By default, this method will wait for a CmdResult message, but you can provide
        any other filtering function/lambda if you are expecting another message as a
        reply from the device.

        :param Message message: WHAD message to send to the device
        :param filter: Filtering function used to match the expected response from the device.
        """
        try:
            return self.__device.send_command(message, filter)
        except WhadDeviceError as device_error:
            logger.debug('an error occured while communicating with the WHAD device !')
            self.on_error(device_error)

    def on_disconnection(self):
        """Device has disconnected or been closed.
        """
        pass


    ######################################
    # Packet flow handling
    ######################################

    def send_packet(self, packet):
        """Send packet to our device.
        """
        # Monitor this outgoing packet
        self.monitor_packet_tx(packet)

        # Convert packet into the corresponding message
        msg = self.hub.convert_packet(packet)

        if msg is not None:
            resp = self.send_command(msg, message_filter(CommandResult))
            logger.info('[connector] Command sent, result: %s' % resp)
            if resp is None:
                raise WhadDeviceDisconnected()
            else:
                return isinstance(resp, Success)
        else:
            logger.error('[connector] Packet cannot be converted into the corresponding WHAD message')
            return False


    def wait_for_message(self, timeout=None, filter=None, command=False):
        """Waits for a specific message to be received.

        This method reads the message queue and return the first message that matches the
        provided filter. A timeout can be specified and will cause this method to return
        None if this timeout is reached.
        """
        return self.__device.wait_for_message(timeout=timeout, filter=filter, command=command)

    # Message callbacks
    def on_any_msg(self, message):
        """Callback function to process any incoming messages.

        This method MAY be overriden by inherited classes.

        :param message: WHAD message
        """
        pass

    def on_discovery_msg(self, message):
        """Callback function to process incoming discovery messages.

        This method MUST be overriden by inherited classes.

        :param message: Discovery message
        """
        logger.error('method `on_discovery_msg` must be implemented in inherited classes')
        raise RequiredImplementation()

    def on_generic_msg(self, message):
        """Callback function to process incoming generic messages.

        This method MUST be overriden by inherited classes.

        :param message: Generic message
        """
        logger.error('method `on_generic_msg` must be implemented in inherited classes')
        raise RequiredImplementation()

    def on_domain_msg(self, domain, message):
        """Callback function to process incoming domain-related messages.

        This method MUST be overriden by inherited classes.

        :param message: Domain message
        """
        logger.error('method `on_domain_msg` must be implemented in inherited classes')
        raise RequiredImplementation()


    def on_packet(self, packet):
        """Callback function to process incoming packets.

        This method MUST be overriden by inherited classes.

        :param packet: Packet
        :type packet: :class:`scapy.packet.Packet`
        """
        logger.error('method `on_packet` must be implemented in inherited classes')
        raise RequiredImplementation()

    def on_event(self, event):
        """Callback function to process incoming events.

        This method MUST be overriden by inherited classes.

        :param event: Event to process
        :type event: :class:`whad.hub.events.AbstractEvent`
        """
        logger.error("Class: %s" % self.__class__)
        logger.error('method `on_event` must be implemented in inherited classes')
        raise RequiredImplementation()

class BridgeIfaceWrapper(WhadDeviceConnector):
    def __init__(self, device, processor):
        super().__init__(device)
        self.__processor = processor

    def send_message(self, message):
        logger.debug('[PacketProcIfaceWrapper] send_message: %s' % message)
        return super().send_message(message)

    def on_disconnection(self):
        """Notify bridge on disconnection.
        """
        logger.debug('[PacketProcIfaceWrapper] on_disconnection')
        self.__processor.on_disconnect(self)

    def on_any_msg(self, message):
        logger.debug('[PacketProcIfaceWrapper] on_any_msg: %s' % message)
        self.__processor.on_any_msg(self, message)

    def on_generic_msg(self, message):
        #logger.debug('[PacketProcIfaceWrapper] on_generic_msg: %s' % message)
        #self.__processor.on_other_msg(self, message)
        pass

    def on_discovery_msg(self, message):
        #logger.debug('[PacketProcIfaceWrapper] on_discovery_msg: %s' % message)
        #self.__processor.on_other_msg(self, message)
        pass

    def on_domain_msg(self, domain, message):
        #self.__processor.on_other_msg(self, message)
        pass

    def on_packet(self, packet):
        #self.__processor.on_packet(self, packet)
        pass

    def on_event(self, event):
        #self.__processor.on_event(self, event)
        pass


class Bridge(object):

    def __init__(self, input_connector, output_connector):
        """Initialize our packet processor.
        """
        # Save our input connector and its device
        self.__in = input_connector
        self.__in_device = self.__in.device

        # Save our output cnnector and its device
        self.__out = output_connector
        self.__out_device =self.__out.device

        # Disable message queue filters
        self.__in_device.set_queue_filter(None)
        self.__out_device.set_queue_filter(None)

        # Bridge our two interfaces with our own connectors.
        # This will replace each device connector with our own and avoid
        # any other packet processing.
        self.__in_wrapper = BridgeIfaceWrapper(self.__in_device, self)
        self.__out_wrapper = BridgeIfaceWrapper(self.__out_device, self)

    def detach(self):
        """Detach BridgeIfaceWrappers from bridge's devices.
        """
        self.__in_device.set_connector(self.__in)
        self.__out_device.set_connector(self.__out)


    @property
    def input(self):
        return self.__in

    @property
    def output(self):
        return self.__out

    def on_disconnect(self, wrapper):
        """When a wrapper disconnects, stop bridge.
        """
        pass

    def on_any_msg(self, wrapper, message):
        if wrapper == self.__in_wrapper:
            self.on_outbound(message)
        else:
            self.on_inbound(message)

    def on_inbound(self, message):
        """Inbound message hook

        This hook is called whenever a message is received on the input
        connector and about to be relayed to the output connector.

        :param message: Message to process
        :type message: HubMessage
        """
        if message is not None:
            # Monitor packet if required
            if issubclass(message, AbstractPacket):
                packet = message.to_packet()
                self.__in.monitor_packet_rx(packet)
                self.__out.monitor_packet_tx(packet)

            # Forward message
            self.__in.send_message(message)

    def on_outbound(self, message):
        """Outbound message hook.

        This hook is called whenever a message is received on the output
        connector and about to be relayed to the input connector.

        :param message: Message to process
        :type message: HubMessage
        """
        if message is not None:
            # Monitor packet if required
            if issubclass(message, AbstractPacket):
                packet = message.to_packet()
                self.__out.monitor_packet_tx(packet)
                self.__in.monitor_packet_rx(packet)

            self.__out.send_message(message)

class WhadDeviceInputThread(Thread):

    """WhadDevice I/O cancellable Thread.

    This thread runs in background and regularly calls the
    device read() method to fetch any incoming data.
    """

    def __init__(self, device, io_thread):
        super().__init__()
        self.__device = device
        self.__io_thread = io_thread
        self.__canceled = False

    def cancel(self):
        """
        Cancel current I/O task.
        """
        self.__canceled = True

    def run(self):
        """
        Main task, call device process() method until thread
        is canceled.
        """
        while not self.__canceled:
            try:
                self.__device.read()
                #logger.debug("[WhadDeviceInputThread::run()] Read data from device")
            except WhadDeviceNotReady as not_ready:
                logger.debug('Device %s has just disconnected (device not ready)' % self.__device.interface)
                break
            except WhadDeviceDisconnected as err:
                logger.debug('Device %s has just disconnected (read returned None)' % self.__device.interface)
                break
        logger.info('Device IO thread canceled and stopped.')
        self.__io_thread.on_disconnection()
        logger.info('on_disconnection done.')

class WhadDeviceMessageThread(Thread):

    def __init__(self, device):
        super().__init__()
        self.__device = device
        self.__canceled = False

    def cancel(self):
        """
        Cancel current I/O task.
        """
        self.__canceled = True

    def run(self):
        """
        Main task, call device process_messages() method until thread
        is canceled.
        """
        while not self.__canceled:
            self.__device.process_messages()

        # Finish processing remaining messages
        logger.debug('[WhadDeviceMessageThread] processing remaining messages ...')
        while self.__device.process_messages(timeout=.1):
            pass

        logger.info('Device message thread canceled and stopped, closing device %s.', self.__device)
        self.__device.close()

class WhadDeviceIOThread(object):

    def __init__(self, device):
        self.__device = device
        self.__input = WhadDeviceInputThread(device, self)
        self.__input.daemon = True
        self.__processing = WhadDeviceMessageThread(device)
        self.__processing.daemon = True
        self.__disconnected = False
        self.__alive = False

    def is_alive(self) -> bool:
        """Check if IO thread is still alive.
        """
        return self.__alive

    def cancel(self):
        self.__input.cancel()
        self.__processing.cancel()

    def start(self):
        logger.info('starting WhadDevice IO management thread ...')
        self.__input.start()
        self.__processing.start()
        logger.info('WhadDevice IO management thread up and running.')
        self.__alive = True

    def join(self):
        logger.info('waiting for WhadDevice IO management thread to finish ...')
        while not self.__disconnected: #and (self.__input.is_alive() or self.__processing.is_alive()):
            try:
                logger.info("Waiting for io thread ...")
                self.__input.join(1.0)
            except RuntimeError:
                logger.debug("RuntimeError raised while joining input thread, we may try to wait for our thread to finish :/")
                pass
            try:
                logger.info("Waiting for processing thread ...")
                self.__processing.join(1.0)
            except RuntimeError:
                logger.debug("RuntimeError raised while joining processing thread, we may try to wait for our thread to finish :/")
                pass

        logger.info('WhadDevice IO management thread finished.')
        self.__alive = False
        if self.__device.opened:
            logger.info('Closing device due to IO termination.')
            self.__device.close()

    def on_disconnection(self):
        """Handle underlying device disconnection
        """
        logger.debug('[device::io_thread] Adapter has disconnected')
        self.__disconnected = True
        self.__input.cancel()
        #self.__device.close()
        self.__processing.cancel()


class WhadDevice(object):
    """
    WHAD Device interface class.

    This device class handles the device discovery process, every possible
    discovery and generic messages related to the device discovery. It MUST be
    inherited by device handling classes (such as the UartDevice class) in order
    to provide read/write capabilities.

    Inherited classes MUST only implement the following methods:

      * open(): will handle device opening/access
      * close(): will handle device closing
      * read() to read data from the device and send them to on_data_received()
      * write() to send data to the device

    All the message re-assembling, parsing, dispatching and background data reading
    will be performed in this class.
    """

    @classmethod
    def _get_sub_classes(cls):
        """
        Helper allowing to get every subclass of WhadDevice.
        """
        # List every available device class
        device_classes = set()
        for device_class in cls.__subclasses__():
            if device_class.__name__ == "VirtualDevice":
                for virtual_device_class in device_class.__subclasses__():
                    device_classes.add(virtual_device_class)
            else:
                device_classes.add(device_class)
        return device_classes

    @classmethod
    def _create(cls, interface_string):
        """
        Helper allowing to get a device according to the interface string provided.

        To make it work, every device class must implement:
            - a class attribute INTERFACE_NAME, matching the interface name
            - a class method list, returning the available devices
            - a property identifier, allowing to identify the device in a unique way

        This method should NOT be used outside of this class. Use WhadDevice.create instead.
        """

        if interface_string.startswith(cls.INTERFACE_NAME):
            identifier = None
            index = None
            if len(interface_string) == len(cls.INTERFACE_NAME):
                index = 0
            elif interface_string[len(cls.INTERFACE_NAME)] == ":":
                index = None
                try:
                    _, identifier = interface_string.split(":")
                except ValueError:
                    identifier = None
            else:
                try:
                    index = int(interface_string[len(cls.INTERFACE_NAME):])
                except ValueError:
                    index = None

            available_devices = cls.list()
            # If the list of device is built statically, check before instantiation
            if available_devices is not None:
                if index is not None:
                    try:
                        return available_devices[index]
                    except IndexError:
                        raise WhadDeviceNotFound
                elif identifier is not None:
                    for dev in available_devices:
                        if dev.identifier == identifier:
                            return dev
                    raise WhadDeviceNotFound
                else:
                    raise WhadDeviceNotFound
            # Otherwise, check dynamically using check_interface
            else:
                formatted_interface_string = interface_string.replace(
                    cls.INTERFACE_NAME + ":",
                    ""
                )
                if cls.check_interface(formatted_interface_string):
                    return cls(formatted_interface_string)
                raise WhadDeviceNotFound

        else:
            raise WhadDeviceNotFound

    @classmethod
    def create(cls, interface_string):
        '''
        Allows to get a specific device according to the provided interface string.
        The interface string is formed as follow:

        "<device_type>[device_index][:device_identifier]"

        Examples:
            * Instantiating the first available UartDevice:
                "uart" or "uart0"

            * Instantiating the second available UartDevice:
                "uart1"

            * Instantiating an UartDevice linked to /dev/ttyACM0:
                "uart:/dev/ttyACM0"

            * Instantiating the first available UbertoothDevice:
                "ubertooth" or "ubertooth0"

            * Instantiating an UbertoothDevice with serial number "11223344556677881122334455667788":
                ubertooth:11223344556677881122334455667788

        '''
        device_classes = cls._get_sub_classes()

        device = None
        for device_class in device_classes:
            try:
                device = device_class._create(interface_string)
                return device
            except WhadDeviceNotFound:
                continue

        raise WhadDeviceNotFound

    @classmethod
    def list(cls):
        '''
        Returns every available compatible devices.
        '''
        device_classes = cls._get_sub_classes()

        available_devices = []
        for device_class in device_classes:
            device_class_list = device_class.list()
            if device_class_list is not None:
                for device in device_class_list:
                    available_devices.append(device)
        return available_devices

    @classmethod
    def check_interface(cls, interface):
        '''
        Checks dynamically if the device can be instantiated.
        '''
        return False

    @property
    def interface(self):
        '''
        Returns the current interface of the device.
        '''
        if hasattr(self.__class__,"INTERFACE_NAME"):
            return self.__class__.INTERFACE_NAME + str(self.index)
        else:
            return "unknown"

    @property
    def type(self):
        '''
        Returns the name of the class linked to the current device.
        '''
        return self.__class__.__name__


    @property
    def opened(self):
        return self.__opened

    def __init__(self):
        # Device information
        self.__info = None
        self.__discovered = False
        self.__opened = False
        self.__closing = False

        # Generate device index
        self.inc_dev_index()
        self.__index = self.__class__.CURRENT_DEVICE_INDEX

        # Device connectors
        self.__connector = None

        # Device IO thread
        self.__io_thread = None

        # Default timeout for messages (5 seconds)
        self.__timeout = 5.0

        # Message queues
        self.__messages = Queue()
        self.__msg_queue = Queue()
        self.__mq_filter = None

        # Input pipes
        self.__inpipe = bytearray()

        # Create locks
        self.__lock = Lock()
        self.__tx_lock = Lock()

        # Protocol hub
        self.__hub = ProtocolHub(2)

    @property
    def hub(self):
        """Retrieve the device protocol hub (parser/factory)
        """
        return self.__hub

    @property
    def index(self):
        return self.__index

    @classmethod
    def inc_dev_index(cls):
        """Inject and maintain device index.
        """
        if hasattr(cls, 'CURRENT_DEVICE_INDEX'):
            cls.CURRENT_DEVICE_INDEX += 1
        else:
            cls.CURRENT_DEVICE_INDEX = 0

    def lock(self):
        """Locks the pending output data buffer."""
        self.__lock.acquire()

    def unlock(self):
        """Unlocks the pending output data buffer."""
        self.__lock.release()

    def set_connector(self, connector):
        """
        Set this device connector.

        :param WhadDeviceConnector connector: connector to be used with this device.
        """
        self.__connector = connector

    ######################################
    # Device I/O operations
    ######################################

    def open(self):
        """
        Open device method. By default, creates a simple thread
        that will handle I/O in background. This requires the object
        to be ready for I/O operations when this method is called.

        This method MUST be overriden by inherited classes.
        """
        self.__io_thread = WhadDeviceIOThread(self)
        self.__io_thread.start()

        # Ask firmware for a reset
        try:
            logger.info('resetting device (if possible)')
            self.__opened = True
            self.reset()
        except Empty as err:
            # Device is unresponsive, does not seem compatible
            # Shutdown IO thread
            self.__io_thread.cancel()
            self.__io_thread.join()

            # Notify device not found
            raise WhadDeviceNotReady()

    def close(self):
        """
        Close device.

        This method MUST be overriden by inherited classes.
        """
        # Avoid recursion when closing
        if not self.__opened or self.__closing:
            logger.debug('exiting close() to avoid recursion')
            return

        logger.info('closing WHAD device')
        self.__closing = True

        # Cancel I/O thread if required
        if self.__io_thread is not None:
            if self.__io_thread.is_alive():
                self.__io_thread.cancel()

        # Send a NOP message to unlock process_messages()
        logger.debug('send NOP message')
        msg = self.hub.generic.create_verbose(b'')
        self.on_message_received(msg)
        logger.debug('NOP message sent')

        # Wait for the thread to terminate nicely.
        if self.__io_thread is not None:
            self.__io_thread.join()

        self.__opened = False
        self.__closing = False

        # Notify connector device has closed
        if self.__connector is not None:
            self.__connector.on_disconnection()

    def wait(self):
        logger.debug("[WhadDevice] waiting for background thread to gracefully stop")
        self.__io_thread.join()
        logger.debug("[WhadDevice] background threads stopped")

    def is_open(self):
        """Determine if the device has been opened or not.
        """
        return self.__opened

    def __write(self, data):
        """
        Sends data to the device.

        This is an internal method that SHALL NOT be used from inherited classes.
        """
        self.lock()
        logger.debug('sending %s to WHAD device %s' % (bytes(data), self.interface))
        self.write(bytes(data))
        self.unlock()


    def set_queue_filter(self, filter=None):
        """Sets the message queue filter.

        :param filter: filtering function/lambda to be used by our message queue filter.
        """
        logger.debug('set queue filter: %s' % filter)
        self.__mq_filter = filter

    def wait_for_single_message(self, timeout, filter=None):
        """Configures the device message queue filter to automatically move messages
        that matches the filter into the queue, and then waits for the first message
        that matches this filter and returns it.
        """
        if filter is not None:
            self.set_queue_filter(filter)

        # Wait for a matching message to be caught (blocking)
        return self.__msg_queue.get(block=True, timeout=timeout)


    def wait_for_message(self, timeout=None, filter=None, command=False):
        """
        Configures the device message queue filter to automatically move messages
        that matches the filter into the queue, and then waits for the first message
        that matches this filter and process it.

        This method is blocking until a matching message is received.

        :param int timeout: Timeout
        :param filter: Message queue filtering function (optional)
        """

        # Check if device is still opem
        if not self.__opened and self.__msg_queue.empty():
            raise WhadDeviceDisconnected()

        logger.debug('entering wait_for_message ...')
        if filter is not None:
            self.set_queue_filter(filter)

        start_time = time()

        while True:
            try:
                # Wait for a matching message to be caught (blocking)
                msg = self.__msg_queue.get(block=True, timeout=timeout)

                # If message does not match, dispatch.
                if not self.__mq_filter(msg):
                    self.dispatch_message(msg)
                    logger.debug('exiting wait_for_message ...')
                else:
                    logger.debug('exiting wait_for_message ...')
                    return msg
            except Empty as err:
                """
                Queue is empty, wait for a message to show up.
                """
                if timeout is not None and (time() - start_time > timeout):
                    if command:
                        raise WhadDeviceTimeout('WHAD device did not answer to a command')
                    else:
                        logger.debug('exiting wait_for_message ...')
                        return None

                sleep(0.001)

        logger.debug('exiting wait_for_message ...')


    def send_message(self, message, keep=None):
        """
        Serializes a message and sends it to the device, without waiting for an answer.
        Optionally, you can update the message queue filter if you need to wait for
        specific messages after the message sent.

        :param Message message: Message to send
        :param keep: Message queue filter function
        """
        logger.info('sending message %s to device <%s>' % (message, self.interface))

        # if `keep` is set, configure queue filter
        if keep is not None:
            logger.debug('send_message:set_queue_filter')
            self.set_queue_filter(keep)

        # Convert message into bytes
        raw_message = message.serialize()


        # Define header
        header = [
            0xAC, 0xBE,
            len(raw_message) & 0xff,
            (len(raw_message) >> 8) & 0xff
        ]

        # Send header followed by serialized message
        self.__write(header)
        self.__write(raw_message)


    def send_command(self, command, keep=None):
        """
        Sends a command and awaits a specific response from the device.
        WHAD commands usualy expect a CmdResult message, if `keep` is not
        provided then this method will by default wait for a CmdResult.

        :param Message command: Command message to send to the device
        :param keep: Message queue filter function (optional)
        :returns: Response message from the device
        :rtype: Message
        """
        self.__tx_lock.acquire()

        # If a queue filter is not provided, expect a default CmdResult
        try:
            if keep is None:
                self.send_message(command, message_filter(CommandResult))
            else:
                self.send_message(command, keep)
        except WhadDeviceError as error:
            # Device error has been triggered, it looks like our device is in
            # an unspecified state, notify user.
            logger.debug('WHAD device in error while sending message: %s' % error)
            raise error

        try:
            # Retrieve the first message matching our filter.
            result = self.wait_for_message(self.__timeout, command=True)
        except WhadDeviceTimeout as timedout:
            # Ensure tx lock is properly released
            self.__tx_lock.release()

            # Forward exception
            raise timedout

        # Ensure tx lock is properly released
        self.__tx_lock.release()

        # Log message
        logger.debug('Command result: %s' % result)

        return result


    def on_data_received(self, data):
        """
        Data received callback.

        This callback will process incoming messages, parse them
        and then forward to the message processing callback.

        :param bytes data: Data received from the device.
        """
        #logger.info('[WhadDevice] entering on_data_received()')
        logger.debug('[WhadDevice] received raw data from device <%s>: %s' % (self.interface, hexlify(data)))
        messages = []
        self.__inpipe.extend(data)
        while len(self.__inpipe) > 2:
            # Is the magic correct ?
            if self.__inpipe[0] == 0xAC and self.__inpipe[1] == 0xBE:
                # Have we received a complete message ?
                if len(self.__inpipe) > 4:
                    msg_size = self.__inpipe[2] | (self.__inpipe[3] << 8)
                    if len(self.__inpipe) >= (msg_size+4):
                        raw_message = self.__inpipe[4:4+msg_size]

                        # Parse our message with our Protocol Hub
                        #_msg = Message()
                        #_msg.ParseFromString(bytes(raw_message))

                        # Parse received message with our Protocol Hub
                        msg = self.__hub.parse(bytes(raw_message))

                        #logger.debug('[WhadDevice] WHAD message successfully parsed')
                        self.on_message_received(msg)
                        # Chomp
                        self.__inpipe = self.__inpipe[msg_size + 4:]
                        #logger.debug('[WhadDevice] Remaining bytes: %s' % hexlify(self.__inpipe))
                    else:
                        break
                else:
                    break
            else:
                #logger.info('[WhadDevice] incorrect header received !')
                # Nope, that's not a header
                while (len(self.__inpipe) >= 2):
                    if (self.__inpipe[0] != 0xAC) or (self.__inpipe[1] != 0xBE):
                        self.__inpipe = self.__inpipe[1:]
                    else:
                        break
        #logger.info('[WhadDevice] exiting on_data_received()')


    def dispatch_message(self, message):
        """Dispatches an incoming message to the corresponding callbacks depending on its
        type and content.

        :param Message message: Message to dispatch
        """
        logger.info('dispatching WHAD message ...')

        # Allows a connector to catch any message
        self.on_any_msg(message)

        # Forward to dedicated callbacks
        if message.message_type == "discovery":
            logger.info('message is about device discovery, forwarding to discovery handler')
            self.on_discovery_msg(message)
        elif message.message_type == "generic":
            logger.info('message is generic, forwarding to default handler')
            self.on_generic_msg(message)
            logger.info('on_generic_message called')
        else:
            domain = message.message_type
            if domain is not None:
                logger.info('message concerns domain `%s`, forward to domain-specific handler' % domain)
                self.on_domain_msg(domain, message)

    def on_message_received(self, message):
        """
        Method called when a WHAD message is received, dispatching.

        :param Message message: Message received
        """
        if self.__closing:
            return

        logger.debug('[WhadDevice::on_message_received()][%s] message received: %s' % (self.interface, message))
        logger.debug('[WhadDevice::on_message_received()][%s] message queue filter: %s' % (self.interface, self.__mq_filter))
        # If message queue filter is defined and message matches this filter,
        # move it into our message queue.
        if self.__mq_filter is not None and self.__mq_filter(message):
            logger.info('message does match current filter, save it for processing')
            self.__msg_queue.put(message, block=True)
            logger.info('message added to message queue')
        else:
            # Save message for background dispatch
            logger.info('message does not match filter or no filter set, save in default message queue')
            self.__messages.put(message, block=True)
            logger.info('message added to default message queue')

    def process_messages(self, timeout=1.0) -> bool:
        """Process pending messages
        """
        result = False

        #if self.__closing:
        #    return

        try:
            message = self.__messages.get(block=True, timeout=timeout)
            if message is not None:
                logger.debug('[process_messages] retrieved message %s' % message)
                self.dispatch_message(message)
                result = True

            return result
        except Empty:
            return False

    ######################################
    # Any messages
    ######################################

    def on_any_msg(self, message):
        """This callback method is called when any message is received

        :param Message message: WHAD message received
        """
        logger.debug('on_any_msg')
        # Forward message to the connector, if any
        if self.__connector is not None:
            logger.debug('Forward message %s to connector %s' % (message, self.__connector))
            self.__connector.on_any_msg(message)

    ######################################
    # Generic messages handling
    ######################################

    def on_generic_msg(self, message):
        """
        This callback method is called whenever a Generic message is received.

        :param Message message: Generic message received
        """
        # Handle generic result message
        if isinstance(message, CommandResult):
            if message.result_code == CommandResult.UNSUPPORTED_DOMAIN:
                logger.error('domain not supported by this device')
                raise UnsupportedDomain("")

        # Forward everything to the connector, if any
        if self.__connector is not None:
            logger.debug('Forward generic message %s to connector %s' % (message, self.__connector))
            self.__connector.on_generic_msg(message)
        else:
            logger.debug('No connector registered, message lost')


    ######################################
    # Generic discovery
    ######################################

    def on_discovery_msg(self, message):
        """
        Method called when a discovery message is received. If a connector has
        been associated with the device, forward this message to this connector.
        """

        # Forward everything to the connector, if any
        if self.__connector is not None:
            self.__connector.on_discovery_msg(message)

    def has_domain(self, domain):
        """Checks if device supports a specific domain.

        :param Domain domain: Domain
        :returns: True if domain is supported, False otherwise.
        :rtype: bool
        """
        if self.__info is not None:
            return self.__info.has_domain(domain)


    def get_domains(self):
        """Get device' supported domains.

        :returns: list of supported domains
        :rtype: list
        """
        if self.__info is not None:
            return self.__info.domains


    def get_domain_capability(self, domain):
        """Get a device domain capabilities.

        :param Domain domain: Target domain
        :returns: Domain capabilities
        :rtype: DeviceDomainInfoResp
        """
        if self.__info is not None:
            return self.__info.get_domain_capabilities(domain)

    def get_domain_commands(self, domain):
        """Get a device supported domain commands.

        :param Domain domain: Target domain
        :returns: Bitmask of supported commands
        :rtype: int
        """
        if self.__info is not None:
            return self.__info.get_domain_commands(domain)


    def send_discover_info_query(self, proto_version=0x0100):
        """
        Sends a DeviceInfoQuery message and awaits for a DeviceInfoResp
        answer.
        """
        logger.info('preparing a DeviceInfoQuery message')
        msg = self.__hub.discovery.create_info_query(proto_version)
        return self.send_command(
            msg,
            message_filter(InfoQueryResp)
        )


    def send_discover_domain_query(self, domain):
        """
        Sends a DeviceDomainQuery message and awaits for a DeviceDomainResp
        answer.
        """
        logger.info('preparing a DeviceDomainInfoQuery message')
        msg = self.__hub.discovery.create_domain_query(domain)
        return self.send_command(
            msg,
            message_filter(DomainInfoQueryResp)
        )

    def discover(self):
        """
        Performs device discovery (synchronously).

        Discovery process asks the device to provide its description, including
        its supported domains and associated capabilities. For each domain we
        then query the device and get the list of supported commands.
        """
        if not self.__discovered:
            # We send a DeviceInfoQuery message to the device and expect a
            # DeviceInfoResponse in return.
            resp = self.send_discover_info_query()

            # If we have an answer, process it.
            if resp is not None:

                # Ensure response is the one we expect
                assert isinstance(resp, InfoQueryResp)

                # Save device information
                self.__info = WhadDeviceInfo(
                    resp
                )

                # Parse DeviceInfoResponse
                #device_info = self.hub.parse(resp)

                # Update our ProtocolHub version to the device version
                self.__hub = ProtocolHub(resp.proto_min_ver)

                # Query device domains
                logger.info('query supported commands per domain')
                for domain in self.__info.domains:
                    resp = self.send_discover_domain_query(domain)
                    self.__info.add_supported_commands(
                        resp.domain,
                        resp.supported_commands
                    )

                # Mark device as discovered
                logger.info('device discovery done')
                self.__discovered = True

                # Switch to max transport speed
                logger.info('set transport speed to %d' % self.info.max_speed)
                self.change_transport_speed(
                    self.info.max_speed
                )
            else:
                logger.error('device is not ready !')
                raise WhadDeviceNotReady()

    def reset(self):
        """Reset device
        """
        logger.info('preparing a DeviceResetQuery message')
        msg = self.hub.discovery.create_reset_query()
        return self.send_command(
            msg,
            message_filter(DeviceReady)
        )

    def change_transport_speed(self, speed):
        """Set device transport speed.

        Optional.
        """
        pass

    @property
    def device_id(self):
        """Return device ID
        """
        return self.__info.device_id

    @property
    def info(self):
        return self.__info

    ######################################
    # Upper layers (domains) handling
    ######################################

    def on_domain_msg(self, domain, message):
        """
        Callback method handling domain-related messages. Since this layer is not
        managed by the root WhadDevice class, forward it to the upper layer, the
        associated connector (if any).

        :param Domain domain: Target domain
        :param Message message: Domain-related message received
        """

        # Forward everything to the connector, if any
        if self.__connector is not None:

            # Check if message is a received packet
            if issubclass(message, AbstractPacket):
                # Convert message into packet
                packet = message.to_packet()
                if packet is not None:

                    # Report packet to monitors
                    self.__connector.monitor_packet_rx(packet)

                    # Forward to our connector
                    if self.__connector.is_locked():
                        # If we are locked, then add packet to our locked packets
                        self.__connector.add_locked_pdu(packet)
                    elif self.__connector.is_synchronous():
                        # If we are in synchronous mode, add packet as pending
                        self.__connector.add_pending_packet(packet)
                    else:
                        #logger.debug('[WhadDevice] on_domain_msg() for device %s: %s' % (self.interface, message))
                        self.__connector.on_packet(packet)
            elif issubclass(message, AbstractEvent):
                # Convert message into event
                event = message.to_event()
                if event is not None:
                    # Forward to our connector
                    self.__connector.on_event(event)
            else:
                # Forward other messages to on_domain_msg() callback
                self.__connector.on_domain_msg(domain, message)
        return False


# Defines every supported low-level device
from whad.device.uart import UartDevice
from whad.device.tcp import TCPSocketDevice
from whad.device.virtual import VirtualDevice
from whad.device.unix import UnixSocketDevice

__all__ = [
    'WhadDeviceConnector',
    'WhadDeviceInfo',
    'WhadDevice',
    'UartDevice',
    'VirtualDevice',
]



"""
Bluetooth LE Stack Link-layer Manager
"""
from binascii import hexlify
from struct import pack
from random import randint

from threading import Lock

from scapy.layers.bluetooth4LE import *

from whad.common.stack import Layer, alias, source, state, LayerState, instance
from whad.ble.stack.l2cap import L2CAPLayer
from whad.ble.crypto import LinkLayerCryptoManager, generate_random_value, e

import logging
logger = logging.getLogger(__name__)

CONNECTION_UPDATE_REQ = 0x00
CHANNEL_MAP_REQ = 0x01
TERMINATE_IND = 0x02
ENC_REQ = 0x03
ENC_RSP = 0x04
START_ENC_REQ = 0x05
START_ENC_RSP = 0x06
UNKNOWN_RSP = 0x07
FEATURE_REQ = 0x08
FEATURE_RSP = 0x09
PAUSE_ENC_REQ = 0x0A
PAUSE_ENC_RSP = 0x0B
VERSION_IND = 0x0C
REJECT_IND = 0x0D
SLAVE_FEATURE_REQ = 0x0E
CONNECTION_PARAM_REQ = 0x0F
CONNECTION_PARAM_RSP = 0x10
REJECT_IND_EXT = 0x11
PING_REQ = 0x12
PING_RSP = 0x13
LENGTH_REQ = 0x14
LENGTH_RSP = 0x15


''' Kept BleConnection because of pairing stuff in it.

class BleConnection(object):

    def __init__(self, llm, conn_handle, local_peer_addr, remote_peer_addr):
        self.__llm = llm
        self.__conn_handle = conn_handle
        self.__local_peer = local_peer_addr
        self.__remote_peer = remote_peer_addr
        self.__l2cap = BleL2CAP(self)
        self.__encrypted = False
        self.__llcm = None
        self.__version_sent = False
        self.__version_remote = None
        self.__lock = Lock()

        self.__handlers = {
            CONNECTION_UPDATE_REQ: self.on_connection_update_req,
            CHANNEL_MAP_REQ: self.on_channel_map_req,
            TERMINATE_IND: self.on_terminate_ind,
            ENC_REQ: self.on_enc_req,
            ENC_RSP: self.on_enc_rsp,
            START_ENC_REQ: self.on_start_enc_req,
            START_ENC_RSP: self.on_start_enc_rsp,
            UNKNOWN_RSP: self.on_unknown_rsp,
            FEATURE_REQ: self.on_feature_req,
            FEATURE_RSP: self.on_feature_rsp,
            PAUSE_ENC_REQ: self.on_pause_enc_req,
            PAUSE_ENC_RSP: self.on_pause_enc_rsp,
            VERSION_IND: self.on_version_ind,
            REJECT_IND: self.on_reject_ind,
            SLAVE_FEATURE_REQ: self.on_slave_feature_req,
            CONNECTION_PARAM_REQ: self.on_connection_param_req,
            CONNECTION_PARAM_RSP: self.on_connection_param_rsp,
            REJECT_IND_EXT: self.on_reject_ind_ext,
            PING_REQ: self.on_ping_req,
            PING_RSP: self.on_ping_rsp,
            LENGTH_REQ: self.on_length_req,
            LENGTH_RSP: self.on_length_rsp
        }

    @property
    def remote_peer(self):
        return self.__remote_peer

    @property
    def local_peer(self):
        return self.__local_peer

    def lock(self):
        """Lock connection
        """
        self.__lock.acquire()

    def unlock(self):
        """Unlock connection
        """
        self.__lock.release()

    def on_disconnect(self):
        """Connection has been closed.
        """
        # Notify GATT layer that the connection has been terminated.
        if self.__l2cap.gatt is not None:
            self.__l2cap.gatt.on_terminated()
            #self.__llm.on_disconnected()

    def on_ctrl_pdu(self, control):
        """Handle Control PDU at connection-level"""
        ctrl = control.getlayer(BTLE_CTRL)
        if ctrl.opcode in self.__handlers:
            self.__handlers[int(ctrl.opcode)](ctrl.getlayer(1))


    def on_l2cap_data(self, data, fragment=False):
        """Forward L2CAP data to L2CAP layer"""
        self.__l2cap.on_data_received(data, fragment)

    def send_l2cap_data(self, data, fragment=False, encrypt=None):
        """Sends data back
        """
        self.__llm.send_data(self.__conn_handle, data, fragment, encrypt=encrypt)

    def send_control(self, pdu, encrypt=None):
        """Sends back a control PDU
        """
        self.__llm.send_control(self.__conn_handle, pdu, encrypt=encrypt)

    @property
    def gatt_class(self):
        return self.__llm.gatt_class

    @property
    def gatt(self):
        return self.__l2cap.att.gatt

    @property
    def conn_handle(self):
        return self.__conn_handle

    @property
    def remote_version(self):
        return self.__version_remote

    def set_stk(self, stk):
        self.__encrypted = True
        self.__stk = stk

    def set_ltk(self, ltk):
        self.__ltk = ltk

    ### Link-layer control PDU callbacks

    def on_unsupported_opcode(self, opcode):
        self.send_control(
            BTLE_CTRL() / LL_UNKNOWN_RSP(code=opcode)
        )

    def on_connection_update_req(self, conn_update):
        """Connection update is not supported yet
        """
        self.on_unsupported_opcode(CONNECTION_UPDATE_REQ)

    def on_channel_map_req(self, channel_map):
        """Channel map update is not supported yet
        """
        self.on_unsupported_opcode(CHANNEL_MAP_REQ)

    def on_terminate_ind(self, terminate):
        """Terminate this connection
        """
        # Notify Link-layer manager that our connection has been terminated
        self.__llm.on_disconnect(self.__conn_handle)

    def on_enc_req(self, enc_req):
        """Encryption request handler
        """
        # Allowed if we have already negociated an STK
        if self.__stk is not None:

            # Generate our SKD and IV
            self.__skd = randint(0, 0x10000000000000000)
            self.__iv = randint(0, 0x100000000)

            logger.info('[llm] Received LL_ENC_REQ: rand=%s ediv=%s skd=%s iv=%s' % (
                hexlify(pack('<Q', enc_req.rand)),
                hexlify(pack('<H', enc_req.ediv)),
                hexlify(pack('<Q', enc_req.skdm)),
                hexlify(pack('<I', enc_req.ivm)),
            ))

            logger.info('[llm] Initiate connection LinkLayerCryptoManager')

            # Save master rand/iv
            self.__randm = enc_req.rand
            self.__ediv = enc_req.ediv

            # Initiate LLCM
            self.__llcm = LinkLayerCryptoManager(
                self.__stk,
                enc_req.skdm,
                enc_req.ivm,
                self.__skd,
                self.__iv
            )

            # Compute session key
            master_skd = pack(">Q", enc_req.skdm)
            master_iv = pack("<L", enc_req.ivm)
            slave_skd = pack(">Q", self.__skd)
            slave_iv = pack("<L", self.__iv)

            # Generate session key diversifier
            skd = slave_skd + master_skd

            # Generate initialization vector
            iv = master_iv + slave_iv

            # Generate session key
            session_key = e(self.__stk, skd)

            logger.info('[llm] master  skd: %s' % hexlify(master_skd))
            logger.info('[llm] master   iv: %s' % hexlify(master_iv))
            logger.info('[llm] slave   skd: %s' % hexlify(slave_skd))
            logger.info('[llm] slave    iv: %s' % hexlify(slave_iv))
            logger.info('[llm] Session  TK: %s' % hexlify(self.__stk))
            logger.info('[llm] Session  iv: %s' % hexlify(iv))
            logger.info('[llm] Exp. Ses iv: %s' % hexlify(self.__llcm.iv))
            logger.info('[llm] Session key: %s' % hexlify(session_key))

            logger.info('[llm] Send LL_ENC_RSP: skd=%s iv=%s' % (
                hexlify(pack('<Q', self.__skd)),
                hexlify(pack('<I', self.__iv))
            ))

            # Send back our parameters
            self.send_control(
                BTLE_CTRL() / LL_ENC_RSP(
                    skds = self.__skd,
                    ivs = self.__iv
                )
            )

            # Notify encryption enabled
            if not self.__llm.set_encryption(
                self.conn_handle,
                enabled = True,
                key=session_key,
                iv=iv
            ):
                logger.info('[llm] Cannot enable encryption')
            else:
                logger.info('[llm] Encryption enabled in hardware')

            # Start encryption (STK as LTK)
            self.send_control(
                BTLE_CTRL() / LL_START_ENC_REQ(),
                encrypt=False
            )

        else:
            self.send_control(


                BTLE_CTRL() / LL_REJECT_IND(
                    code=0x1A # Unsupported Remote Feature
                )
            )

    def on_enc_rsp(self, enc_rsp):
        """Encryption not supported yet
        """
        self.on_unsupported_opcode(ENC_RSP)

    def on_start_enc_req(self, start_enc_req):
        """Encryption not supported yet
        """
        self.on_unsupported_opcode(START_ENC_REQ)

    def on_start_enc_rsp(self, start_enc_rsp):
        """Encryption start response handler

        Normally, we get this packet when a link has successfully
        been encrypted (with STK or LTK). So we need to notify the
        SMP that encryption has been acknowledged by the remote peer.


        """
        # Check if we are the encryption initiator,
        # if yes then we need to answer to this encrypted LL_START_ENC_RSP
        # with another encrypted LL_START_ENC_RSP
        if not self.__l2cap.smp.is_initiator():
            self.send_control(
                BTLE_CTRL() / LL_START_ENC_RSP()
            )

        # Notify SMP channel is now encrypted
        self.__l2cap.smp.on_channel_encrypted()

    def on_unknown_rsp(self, unk_rsp):
        pass

    def on_feature_req(self, feature_req):
        """Features not supported yet
        """
        #self.on_unsupported_opcode(FEATURE_REQ)
        # Reply with our basic feature set
        self.send_control(
            BTLE_CTRL() / LL_FEATURE_RSP(feature_set=[
                'le_encryption',
                'le_ping'
            ])
        )

    def on_feature_rsp(self, feature_rsp):
        """Features not supported yet
        """
        self.on_unsupported_opcode(FEATURE_RSP)

    def on_pause_enc_req(self, pause_enc_req):
        """Encryption not supported yet
        """
        self.on_unsupported_opcode(PAUSE_ENC_REQ)

    def on_pause_enc_rsp(self, pause_enc_rsp):
        """Encryption not supported yet
        """
        self.on_unsupported_opcode(PAUSE_ENC_RSP)

    def on_version_ind(self, version):
        """Send back our version info
        """
        if not self.__version_sent:
            self.send_control(
                BTLE_CTRL() / LL_VERSION_IND(
                    version=self.__llm.stack.bt_version,
                    company=self.__llm.stack.manufacturer_id,
                    subversion=self.__llm.stack.bt_sub_version
                )
            )
        self.__version_remote = version

    def on_reject_ind(self, reject):
        pass

    def on_slave_feature_req(self, feature_req):
        self.on_unsupported_opcode(FEATURE_REQ)

    def on_connection_param_req(self, conn_param_req):
        self.on_unsupported_opcode(CONNECTION_PARAM_REQ)

    def on_connection_param_rsp(self, conn_param_rsp):
        pass

    def on_reject_ind_ext(self, reject_ext):
        pass

    def on_ping_req(self, ping_req):
        pass

    def on_ping_rsp(self, ping_rsp):
        pass

    def on_length_req(self, length_req):
        """Received a length request PDU
        """
        pass

    def on_length_rsp(self, length_rsp):
        pass

    ##################################
    # LLM control procedures
    ##################################

    def send_version(self):
        """Send LL_VERSION_IND PDU.
        """
        if not self.__version_sent:
            # Mark version as sent
            self.__version_sent = True

            # Send LL_VERSION_IND PDU
            self.send_control(
                BTLE_CTRL() / LL_VERSION_IND(
                    version=self.__llm.stack.bt_version,
                    company=self.__llm.stack.manufacturer_id,
                    subversion=self.__llm.stack.bt_sub_version
                )
            )
'''

class BleConnection(object):

    def __init__(self, l2cap_instance, conn_handle, local_peer_addr, remote_peer_addr):
        self.__l2cap = l2cap_instance
        self.__conn_handle = conn_handle
        self.__remote_peer = remote_peer_addr
        self.__local_peer = local_peer_addr
        self.__version_sent = False
        self.__lock = Lock()

    @property
    def remote_peer(self):
        return self.__remote_peer

    @property
    def local_peer(self):
        return self.__local_peer

    @property
    def conn_handle(self):
        return self.__conn_handle

    @property
    def l2cap(self):
        return self.__l2cap

    @property
    def gatt(self):
        return self.__l2cap.get_layer('gatt')

    @property
    def smp(self):
        return self.__l2cap.get_layer('smp')

    @property
    def phy(self):
        return self.__l2cap.get_layer('phy')

    @property
    def ll(self):
        return self.__l2cap.get_layer('ll')

    @property
    def remote_version(self):
        return self.ll.state.get_version_remote(self.__conn_handle)

    def lock(self):
        """Lock connection
        """
        self.__lock.acquire()

    def unlock(self):
        """Unlock connection
        """
        self.__lock.release()

    def on_disconnect(self):
        """Connection has been closed.
        """
        pass

    def send_version(self):
        """Send LL_VERSION_IND PDU.
        """
        if not self.ll.state.get_connection(self.__conn_handle)['version_sent']:
            # Mark version as sent
            self.ll.state.get_connection(self.__conn_handle)['version_sent'] = True

            # Send LL_VERSION_IND PDU
            self.ll.send_ctrl_pdu(
                self.__conn_handle,
                LL_VERSION_IND(
                    version=self.phy.bt_version.value,
                    company=self.phy.manufacturer_id,
                    subversion=self.phy.bt_sub_version
                )
            )

    def on_version_ind(self, version):
        """Send back our version info
        """
        if not self.__version_sent:
            self.send_control(
                BTLE_CTRL() / LL_VERSION_IND(
                    version=self.__llm.stack.bt_version,
                    company=self.__llm.stack.manufacturer_id,
                    subversion=self.__llm.stack.bt_sub_version
                )
            )
        self.__version_remote = version

class LinkLayerState(LayerState):

    def __init__(self):
        super().__init__()
        self.connections = {}

    def get_connection(self, conn_handle):
        if conn_handle in self.connections:
            return self.connections[conn_handle]
        else:
            raise IndexError

    def register_connection(self, conn_handle, l2cap_instance, local_peer_addr, remote_peer_addr):
        self.connections[conn_handle] = {
            'l2cap': l2cap_instance,
            'local_peer_addr': local_peer_addr.value,
            'local_peer_addr_type': local_peer_addr.type,
            'remote_peer_addr': remote_peer_addr.value,
            'remote_peer_addr_type': remote_peer_addr.type,
            'version_sent': False,  # version exchanged
            'version_remote': None,
            'encryption_key':None,
            'authenticated':False,
            'encrypted':False,
            'skd':None,
            'iv':None,
            'rand':None,
            'ediv':None,
            'nb_pdu_recvd': 0       # number of packets received
        }

    def unregister_connection(self, conn_handle):
        if conn_handle in self.connections:
            del self.connections[conn_handle]

    def get_connection_l2cap(self, conn_handle):
        if conn_handle in self.connections:
            return self.connections[conn_handle]['l2cap']
        return None

    def get_connection_handle(self, l2cap_instance):
        for conn_handle in self.connections:
            if self.connections[conn_handle]['l2cap'] == l2cap_instance:
                return conn_handle
        return

    def register_encryption_key(self, conn_handle, key):
        if conn_handle in self.connections:
            self.connections[conn_handle]['encryption_key'] = key

    def is_authenticated(self, conn_handle):
        if conn_handle in self.connections:
            return self.connections[conn_handle]['authenticated']

    def is_encrypted(self, conn_handle):
        if conn_handle in self.connections:
            return self.connections[conn_handle]['encrypted']

    def mark_as_authenticated(self, conn_handle):
        if conn_handle in self.connections:
            self.connections[conn_handle]['authenticated'] = True

    def mark_as_encrypted(self, conn_handle):
        if conn_handle in self.connections:
            self.connections[conn_handle]['encrypted'] = True

    def get_encryption_key(self, conn_handle):
        if conn_handle in self.connections:
            return self.connections[conn_handle]['encryption_key']
        return None

    def register_skd_and_iv(self, conn_handle, skd, iv):
        if conn_handle in self.connections:
            self.connections[conn_handle]['skd'] = skd
            self.connections[conn_handle]['iv'] = iv


    def get_skd_and_iv(self, conn_handle):
        if conn_handle in self.connections:
            skd = self.connections[conn_handle]['skd']
            iv = self.connections[conn_handle]['iv']
            return (skd, iv)
        return (None, None)

    def register_rand_and_ediv(self, conn_handle, rand, ediv):
        if conn_handle in self.connections:
            self.connections[conn_handle]['rand'] = rand
            self.connections[conn_handle]['ediv'] = ediv

    def get_rand_and_ediv(self, conn_handle):
        if conn_handle in self.connections:
            rand = self.connections[conn_handle]['rand']
            ediv = self.connections[conn_handle]['ediv']
            return (rand, ediv)
        return (None, None)

    def mark_version_sent(self, conn_handle):
        if conn_handle in self.connections:
            self.connections[conn_handle]['version_sent'] = True

    def is_version_sent(self, conn_handle):
        if conn_handle in self.connections:
            return self.connections[conn_handle]['version_sent']
        return False

    def set_version_remote(self, conn_handle, version):
        if conn_handle in self.connections:
            self.connections[conn_handle]['version_remote'] = version

    def get_version_remote(self, conn_handle):
        if conn_handle in self.connections:
            return self.connections[conn_handle]['version_remote']

@alias('ll')
@state(LinkLayerState)
class LinkLayer(Layer):

    def configure(self, options={}):
        # Control PDU dispatch
        self.__handlers = {
            CONNECTION_UPDATE_REQ: self.on_connection_update_req,
            CHANNEL_MAP_REQ: self.on_channel_map_req,
            TERMINATE_IND: self.on_terminate_ind,
            ENC_REQ: self.on_enc_req,
            ENC_RSP: self.on_enc_rsp,
            START_ENC_REQ: self.on_start_enc_req,
            START_ENC_RSP: self.on_start_enc_rsp,
            UNKNOWN_RSP: self.on_unknown_rsp,
            FEATURE_REQ: self.on_feature_req,
            FEATURE_RSP: self.on_feature_rsp,
            PAUSE_ENC_REQ: self.on_pause_enc_req,
            PAUSE_ENC_RSP: self.on_pause_enc_rsp,
            VERSION_IND: self.on_version_ind,
            REJECT_IND: self.on_reject_ind,
            SLAVE_FEATURE_REQ: self.on_slave_feature_req,
            CONNECTION_PARAM_REQ: self.on_connection_param_req,
            CONNECTION_PARAM_RSP: self.on_connection_param_rsp,
            REJECT_IND_EXT: self.on_reject_ind_ext,
            PING_REQ: self.on_ping_req,
            PING_RSP: self.on_ping_rsp,
            LENGTH_REQ: self.on_length_req,
            LENGTH_RSP: self.on_length_rsp
        }

    def on_connect(self, conn_handle, local_peer_addr, remote_peer_addr):
        """Handles BLE connection
        """
        if conn_handle not in self.state.connections:
            logger.info('[llm] registers new connection %d with %s' % (conn_handle, remote_peer_addr))

            # Instantiate a L2CAP layer (contextual) to handle the connection
            conn_l2cap = self.instantiate(L2CAPLayer)
            conn_l2cap.set_conn_handle(conn_handle)

            # Update state with new connection
            self.state.register_connection(conn_handle, conn_l2cap.name, local_peer_addr, remote_peer_addr)

            # Return connection object
            return BleConnection(
                conn_l2cap,
                conn_handle,
                local_peer_addr,
                remote_peer_addr
            )
        else:
            logger.error('[!] Connection already exists')

            # Return connection object
            return BleConnection(
                self,
                conn_handle,
                local_peer_addr,
                remote_peer_addr
            )

    def on_disconnect(self, conn_handle):
        # Free the previously instantiated L2CAP layer
        conn_layer = self.state.get_connection_l2cap(conn_handle)
        if conn_layer is not None:
            self.destroy(self.get_layer(conn_layer))

        # Remove connection from our registered connections
        self.state.unregister_connection(conn_handle)

    @source('phy', 'data')
    def on_data_pdu_recv(self, pdu, conn_handle=None):
        """We received a DATA PDU.
        """
        # Count PDU
        if conn_handle in self.state.connections:
            conn_metadata = self.state.get_connection(conn_handle)
            conn_metadata['nb_pdu_recvd'] += 1

        # We received a data PDU
        self.on_data_pdu(pdu, conn_handle)

    @source('phy', 'control')
    def on_ctrl_pdu_recv(self, pdu, tag=None, conn_handle=None):
        """We received a CTRL PDU.
        """
        # Count PDU
        if conn_handle in self.state.connections:
            conn_metadata = self.state.get_connection(conn_handle)
            conn_metadata['nb_pdu_recvd'] += 1

        # We received a control PDU
        self.on_ctrl_pdu(pdu, conn_handle)

    def on_ctrl_pdu(self, pdu, conn_handle):
        if conn_handle in self.state.connections:
            ctrl = pdu.getlayer(BTLE_CTRL)
            if ctrl.opcode in self.__handlers:
                self.__handlers[int(ctrl.opcode)](conn_handle, ctrl.getlayer(1))
        else:
            logger.error('[!] Unknown connection handle: %d', conn_handle)


    def on_data_pdu(self, pdu, conn_handle):
        """Forward data PDU to upper layer (L2CAP)
        """
        # We look for the corresponding L2CAP layer instance
        l2cap_layer = self.state.get_connection_l2cap(conn_handle)
        if l2cap_layer is not None:
            self.send(l2cap_layer, bytes(pdu.payload), fragment=(pdu.LLID == 0x1))

    @instance('l2cap')
    def on_l2cap_send_data(self, instance, data, fragment=False, encrypt=None):
        '''L2CAP data encapsulation

        This method retrieves the connection handle corresponding to the instance
        that wants to send data and build a BTLE_DATA packet to send to the PHY
        layer.
        '''
        # Retrieve connection handle corresponding to the instance
        conn_handle = self.state.get_connection_handle(instance)
        if conn_handle is not None:
            logger.debug('sending l2cap data PDU for conn_handle %d' % conn_handle)
            llid = 0x01 if fragment else 0x02
            self.send(
                'phy',
                BTLE_DATA(
                    LLID=llid,
                    len=len(data)
                )/data,
                tag='data',
                conn_handle=conn_handle,
                encrypt=encrypt
            )
        else:
            logger.error('no connection handle found for L2CAP instance %s' % instance)

    def send_ctrl_pdu(self, conn_handle, pdu, encrypt=None):
        """Send a control PDU to the underlying PHY layer.
        """
        self.send('phy', BTLE_DATA()/BTLE_CTRL()/pdu, tag='control', conn_handle=conn_handle, encrypt=encrypt)

    """Control PDU handlers
    """

    ### Link-layer control PDU callbacks

    def on_unsupported_opcode(self, conn_handle, opcode):
        self.send_ctrl_pdu(
            conn_handle,
            LL_UNKNOWN_RSP(code=opcode)
        )

    def on_connection_update_req(self, conn_handle, conn_update):
        """Connection update is not supported yet
        """
        self.on_unsupported_opcode(conn_handle, CONNECTION_UPDATE_REQ)

    def on_channel_map_req(self, conn_handle, channel_map):
        """Channel map update is not supported yet
        """
        self.on_unsupported_opcode(conn_handle, CHANNEL_MAP_REQ)

    def on_terminate_ind(self, conn_handle, terminate):
        """Terminate this connection
        """
        # Connection has been terminated
        conn = self.state.get_connection(conn_handle)
        if conn is not None:
            self.on_disconnect(conn_handle)

    def start_encryption(self, conn_handle, rand, ediv):
        """
        Initiate encryption procedure.
        """
        # Retrieve connection handle corresponding to the instance

        encryption_key = None
        if conn_handle is not None:
            encryption_key = self.state.get_encryption_key(conn_handle)

        # Allowed if we have already negociated an STK
        if encryption_key is not None and conn_handle is not None:

            # Generate our SKD and IV
            skd = randint(0, 0x10000000000000000)
            iv = randint(0, 0x100000000)
            self.state.register_skd_and_iv(conn_handle, skd, iv)

            logger.info('[llm] Initiate connection LinkLayerCryptoManager')

            # Save master rand/iv
            self.state.register_rand_and_ediv(conn_handle, rand, ediv)

            # Send back our parameters
            self.send_ctrl_pdu(
                conn_handle,
                LL_ENC_REQ(
                    rand = rand,
                    ediv = ediv,
                    skdm = skd,
                    ivm = iv
                )
            )

        else:
            self.send_ctrl_pdu(
                conn_handle,
                LL_REJECT_IND(
                    code=0x1A # Unsupported Remote Feature
                )
            )

    def on_enc_rsp(self, conn_handle, enc_rsp):
        """Encryption response handler
        """
        # Retrieve connection handle corresponding to the instance

        encryption_key = None
        if conn_handle is not None:
            encryption_key = self.state.get_encryption_key(conn_handle)


        # Allowed if we have already negociated an STK
        if encryption_key is not None and conn_handle is not None:

            skdm, ivm = self.state.get_skd_and_iv(conn_handle)
            rand, ediv = self.state.get_rand_and_ediv(conn_handle)

            logger.info('[llm] Received LL_ENC_RSP: skds=%s ivs=%s' % (
                hexlify(pack('<Q', enc_rsp.skds)),
                hexlify(pack('<I', enc_rsp.ivs)),
            ))

            logger.info('[llm] Initiate connection LinkLayerCryptoManager')

            # Initiate LLCM
            self.__llcm = LinkLayerCryptoManager(
                encryption_key,
                skdm,
                ivm,
                enc_rsp.skds,
                enc_rsp.ivs
            )

            # Compute session key
            master_skd = pack(">Q", skdm)
            master_iv = pack("<L", ivm)
            slave_skd = pack(">Q", enc_rsp.skds)
            slave_iv = pack("<L", enc_rsp.ivs)

            # Generate session key diversifier
            skd = slave_skd + master_skd

            # Generate initialization vector
            iv = master_iv + slave_iv

            # Generate session key
            session_key = e(encryption_key, skd)

            logger.info('[llm] master  skd: %s' % hexlify(master_skd))
            logger.info('[llm] master   iv: %s' % hexlify(master_iv))
            logger.info('[llm] slave   skd: %s' % hexlify(slave_skd))
            logger.info('[llm] slave    iv: %s' % hexlify(slave_iv))
            logger.info('[llm] Session  TK: %s' % hexlify(encryption_key))
            logger.info('[llm] Session  iv: %s' % hexlify(iv))
            logger.info('[llm] Exp. Ses iv: %s' % hexlify(self.__llcm.iv))
            logger.info('[llm] Session key: %s' % hexlify(session_key))


            # Notify encryption enabled
            if not self.get_layer('phy').set_encryption(
                conn_handle=conn_handle,
                enabled=True,
                ll_key=session_key,
                ll_iv=iv,
                key=encryption_key,
                rand=rand,
                ediv=ediv
            ):
                logger.info('[llm] Cannot enable encryption')
            else:
                logger.info('[llm] Encryption enabled in hardware')

        else:
            self.send_ctrl_pdu(
                conn_handle,
                LL_REJECT_IND(
                    code=0x1A # Unsupported Remote Feature
                )
            )


    def on_enc_req(self, conn_handle, enc_req):
        """Encryption request handler
        """
        # Retrieve connection handle corresponding to the instance

        encryption_key = None
        if conn_handle is not None:
            encryption_key = self.state.get_encryption_key(conn_handle)

        # Allowed if we have already negociated an STK
        if encryption_key is not None and conn_handle is not None:

            # Generate our SKD and IV
            skd = randint(0, 0x10000000000000000)
            iv = randint(0, 0x100000000)
            self.state.register_skd_and_iv(conn_handle, skd, iv)

            logger.info('[llm] Received LL_ENC_REQ: rand=%s ediv=%s skd=%s iv=%s' % (
                hexlify(pack('<Q', enc_req.rand)),
                hexlify(pack('<H', enc_req.ediv)),
                hexlify(pack('<Q', enc_req.skdm)),
                hexlify(pack('<I', enc_req.ivm)),
            ))

            logger.info('[llm] Initiate connection LinkLayerCryptoManager')

            # Save master rand/iv
            self.state.register_rand_and_ediv(conn_handle, enc_req.rand, enc_req.ediv)

            # Initiate LLCM
            self.__llcm = LinkLayerCryptoManager(
                encryption_key,
                enc_req.skdm,
                enc_req.ivm,
                skd,
                iv
            )

            # Compute session key
            master_skd = pack(">Q", enc_req.skdm)
            master_iv = pack("<L", enc_req.ivm)
            slave_skd = pack(">Q", skd)
            slave_iv = pack("<L", iv)

            # Generate session key diversifier
            skd = slave_skd + master_skd

            # Generate initialization vector
            iv = master_iv + slave_iv

            # Generate session key
            session_key = e(encryption_key, skd)

            logger.info('[llm] master  skd: %s' % hexlify(master_skd))
            logger.info('[llm] master   iv: %s' % hexlify(master_iv))
            logger.info('[llm] slave   skd: %s' % hexlify(slave_skd))
            logger.info('[llm] slave    iv: %s' % hexlify(slave_iv))
            logger.info('[llm] Session  TK: %s' % hexlify(encryption_key))
            logger.info('[llm] Session  iv: %s' % hexlify(iv))
            logger.info('[llm] Exp. Ses iv: %s' % hexlify(self.__llcm.iv))
            logger.info('[llm] Session key: %s' % hexlify(session_key))
            skdm, ivm = self.state.get_skd_and_iv(conn_handle)
            logger.info('[llm] Send LL_ENC_RSP: skd=%s iv=%s' % (
                hexlify(pack('<Q', skdm)),
                hexlify(pack('<I', ivm))
            ))

            # Send back our parameters
            self.send_ctrl_pdu(
                conn_handle,
                LL_ENC_RSP(
                    skds = skdm,
                    ivs = ivm
                )
            )

            # Notify encryption enabled
            if not self.get_layer('phy').set_encryption(
                conn_handle=conn_handle,
                enabled=True,
                ll_key=session_key,
                ll_iv=iv,
                key=encryption_key,
                rand=enc_req.rand,
                ediv=enc_req.ediv
            ):
                logger.info('[llm] Cannot enable encryption')
            else:
                logger.info('[llm] Encryption enabled in hardware')

            # Start encryption (STK as LTK)
            self.send_ctrl_pdu(
                conn_handle,
                LL_START_ENC_REQ(),
                encrypt=False
            )

        else:
            self.send_ctrl_pdu(
                conn_handle,
                LL_REJECT_IND(
                    code=0x1A # Unsupported Remote Feature
                )
            )

    def on_start_enc_req(self, conn_handle, start_enc_req):
        """Encryption start request handler.
        """

        # Start encryption (STK as LTK)
        self.send_ctrl_pdu(
            conn_handle,
            LL_START_ENC_RSP()
        )

    def on_start_enc_rsp(self, conn_handle, start_enc_rsp):
        """Encryption start response handler

        Normally, we get this packet when a link has successfully
        been encrypted (with STK or LTK). So we need to notify the
        SMP that encryption has been acknowledged by the remote peer.


        """
        # Check if we are the encryption initiator,
        # if yes then we need to answer to this encrypted LL_START_ENC_RSP
        # with another encrypted LL_START_ENC_RSP
        l2cap_instance = self.state.get_connection_l2cap(conn_handle)
        if not self.get_layer(l2cap_instance).get_layer('smp').is_initiator():
            self.send_ctrl_pdu(
                conn_handle,
                LL_START_ENC_RSP()
            )

        #print("Marked as encrypted")
        self.state.mark_as_encrypted(conn_handle)

        # Notify SMP channel is now encrypted
        self.get_layer(l2cap_instance).get_layer('smp').on_channel_encrypted()

    def on_unknown_rsp(self, conn_handle, unk_rsp):
        pass

    def on_feature_req(self, conn_handle, feature_req):
        """Features not supported yet
        """
        #self.on_unsupported_opcode(FEATURE_REQ)
        # Reply with our basic feature set
        self.send_ctrl_pdu(
            conn_handle,
            LL_FEATURE_RSP(feature_set=[
                'le_encryption',
                'le_ping'
            ])
        )

    def on_feature_rsp(self, conn_handle, feature_rsp):
        """Features not supported yet
        """
        self.on_unsupported_opcode(conn_handle, FEATURE_RSP)

    def on_pause_enc_req(self, conn_handle, pause_enc_req):
        """Encryption not supported yet
        """
        self.on_unsupported_opcode(conn_handle, PAUSE_ENC_REQ)

    def on_pause_enc_rsp(self, conn_handle, pause_enc_rsp):
        """Encryption not supported yet
        """
        self.on_unsupported_opcode(conn_handle, PAUSE_ENC_RSP)

    def on_version_ind(self, conn_handle, version):
        """Send back our version info
        """
        logger.debug('received a VERSION_IND PDU')
        if not self.state.is_version_sent(conn_handle):
            logger.debug('sending back our VERSION_IND PDU')

            # send control PDU
            self.send_ctrl_pdu(
                conn_handle,
                LL_VERSION_IND(
                    version=self.get_layer('phy').bt_version.value,
                    company=self.get_layer('phy').manufacturer_id,
                    subversion=self.get_layer('phy').bt_sub_version
                )
            )
        else:
            logger.debug('VERSION_IND PDU already sent, skip.')
        self.state.set_version_remote(conn_handle, version)
        self.state.mark_version_sent(conn_handle)

    def on_reject_ind(self, conn_handle, reject):
        pass

    def on_slave_feature_req(self, conn_handle, feature_req):
        self.on_unsupported_opcode(conn_handle, SLAVE_FEATURE_REQ)

    def on_connection_param_req(self, conn_handle, conn_param_req):
        self.on_unsupported_opcode(conn_handle, CONNECTION_PARAM_REQ)

    def on_connection_param_rsp(self, conn_handle, conn_param_rsp):
        pass

    def on_reject_ind_ext(self, conn_handle, reject_ext):
        pass

    def on_ping_req(self, conn_handle, ping_req):
        self.on_unsupported_opcode(conn_handle, PING_REQ)

    def on_ping_rsp(self, conn_handle, ping_rsp):
        pass

    def on_length_req(self, conn_handle, length_req):
        """Received a length request PDU
        """
        self.on_unsupported_opcode(conn_handle, LENGTH_REQ)

    def on_length_rsp(self, conn_handle, length_rsp):
        pass

LinkLayer.add(L2CAPLayer)



from queue import Queue

from whad.exceptions import WhadDeviceDisconnected
from whad.phy.connector import Phy
from whad.phy import Endianness
from whad.hub.phy import Modulation as PhyModulation, Endianness as PhyEndianness
from whad.phy.injecting import InjectionConfiguration
from whad.phy.exceptions import NoModulation
from whad.exceptions import UnsupportedCapability

# TODO: every sniffer is broken (sniff() method does not catch packets, we
#       have to catch them in on_packet() and put them in a queue)

class Injector(Phy):
    """
    Phy Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device):
        Phy.__init__(self, device)
        self._metadata_check = False
        # Check if device can perform injection
        if not self.can_send():
            raise UnsupportedCapability("Inject")

        self.__configuration = InjectionConfiguration()


    @property
    def configuration(self) -> InjectionConfiguration:
        """Retrieve this injector configuration.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration: InjectionConfiguration):
        """Set the injector configuration.
        """
        self.stop()
        self.__configuration = new_configuration
        if self.__configuration.frequency is not None:
            self.set_frequency(self.__configuration.frequency)
        if self.__configuration.packet_size is not None:
            self.set_packet_size(self.__configuration.packet_size)

        # Set data rate for all modulations but LoRa
        if not self.__configuration.lora:
            if self.__configuration.datarate is not None:
                self.set_datarate(self.__configuration.datarate)

        if self.__configuration.gfsk:
            self.set_gfsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.bfsk:
            self.set_bfsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.qfsk:
            self.set_4fsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.ask:
            self.set_ask()
        elif self.__configuration.bpsk:
            self.set_bpsk()
        elif self.__configuration.qpsk:
            self.set_qpsk()
        elif self.__configuration.lora:
            self.set_lora(
                self.__configuration.lora_configuration.spreading_factor,
                self.__configuration.lora_configuration.coding_rate,
                self.__configuration.lora_configuration.bandwidth,
                12,
                crc=self.__configuration.lora_configuration.enable_crc,
                explicit=self.__configuration.lora_configuration.enable_explicit_mode
            )

        # Set endianness for all modulations but LoRa
        if not self.__configuration.lora:
            self.set_endianness(
                Endianness.LITTLE if
                self.__configuration.little_endian else
                Endianness.BIG
            )
        if self.__configuration.sync_word is not None:
            self.set_sync_word(self.__configuration.sync_word)
        self.start()

    def inject(self, packet):

        if not self._metadata_check:
            if hasattr(packet, "metadata") and hasattr(packet.metadata, "modulation") and packet.metadata.modulation is not None:
                deviation = packet.metadata.deviation
            else:
                deviation = None
            if deviation is not None:
                self.stop()
                if hasattr(packet, "metadata") and hasattr(packet.metadata, "modulation"):
                    if packet.metadata.modulation == PhyModulation.ASK and not self.__configuration.ask:
                        self.set_ask()
                    elif packet.metadata.modulation == PhyModulation.FSK and deviation is not None and not self.__configuration.bfsk:
                        self.set_bfsk(deviation=deviation)
                    elif (packet.metadata.modulation == PhyModulation.GFSK or packet.metadata.modulation == PhyModulation.MSK) and deviation is not None and not self.__configuration.gfsk:
                        self.set_gfsk(deviation=deviation)
                    elif packet.metadata.modulation == PhyModulation.FOURFSK and deviation is not None and not self.__configuration.qfsk:
                        self.set_4fsk(deviation=deviation)
                    elif packet.metadata.modulation == PhyModulation.BPSK and not self.__configuration.bpsk:
                        self.set_bpsk()
                    elif packet.metadata.modulation == PhyModulation.QPSK and not self.__configuration.qpsk:
                        self.set_qpsk()
                    elif packet.metadata.modulation == PhyModulation.LORA and not self.__configuration.lora:
                        self.set_lora(
                            self.__configuration.lora_configuration.spreading_factor,
                            self.__configuration.lora_configuration.coding_rate,
                            self.__configuration.lora_configuration.bandwidth,
                            12,
                            crc=self.__configuration.lora_configuration.enable_crc,
                            explicit=self.__configuration.lora_configuration.enable_explicit_mode
                        )

                if hasattr(packet, "metadata") and hasattr(packet.metadata, "frequency") and self.__configuration.frequency is None:
                    self.set_frequency(packet.metadata.frequency)

                if hasattr(packet, "metadata") and hasattr(packet.metadata, "packet_size") and self.__configuration.packet_size is None:
                    self.set_packet_size(packet.metadata.packet_size)

                if hasattr(packet, "metadata") and hasattr(packet.metadata, "datarate") and self.__configuration.datarate is None:
                    self.set_datarate(packet.metadata.datarate)

                if hasattr(packet, "metadata") and hasattr(packet.metadata, "syncword") and self.__configuration.sync_word is None:
                    self.set_sync_word(packet.metadata.syncword)

                if hasattr(packet, "metadata") and hasattr(packet.metadata, "endianness"):
                    self.set_endianness(
                        Endianness.LITTLE if
                        packet.metadata.endianness == PhyEndianness.LITTLE else
                        Endianness.BIG
                    )

                self.start()
                self._metadata_check = True

        super().send(packet)



from whad.hub.discovery import Domain, Capability
from whad.hub.ble import Commands as BleCommands
from whad.hub.dot15d4 import Commands as Dot15d4Commands
from whad.hub.esb import Commands as EsbCommands
from whad.hub.unifying import Commands as UnifyingCommands
from whad.hub.phy import Commands as PhyCommands

DLT_BLUETOOTH_LE_LL_WITH_PHDR   = 256
DLT_IEEE802_15_4_TAP            = 283
DLT_RESERVED_02                 = 148
DLT_RESERVED_03                 = 149
DLT_RESERVED_06                 = 152

CAPABILITIES = {
    DLT_BLUETOOTH_LE_LL_WITH_PHDR : (
        {
            Domain.BtLE : (
                (Capability.Sniff),
                [BleCommands.SniffConnReq, BleCommands.SniffAdv, BleCommands.Start, BleCommands.Stop]
            )
        },
        {
            Domain.BtLE : (
                (Capability.Inject),
                [BleCommands.SendPDU, BleCommands.Start, BleCommands.Stop]
            )
        }
    ),
    DLT_IEEE802_15_4_TAP : (
        {
            Domain.Dot15d4 : (
                (Capability.Sniff),
                [Dot15d4Commands.Sniff, Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        },
        {
            Domain.Dot15d4 : (
                (Capability.Inject),
                [Dot15d4Commands.Send,Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        }
    ),
    DLT_RESERVED_02 : (
        {
            Domain.Esb : (
                (Capability.Sniff),
                [EsbCommands.Sniff, EsbCommands.Start, EsbCommands.Stop]
            ),
            Domain.LogitechUnifying : (
                (Capability.Sniff),
                [UnifyingCommands.Sniff, UnifyingCommands.Start, UnifyingCommands.Stop]
            )
        },
        {
            Domain.Esb : (
                (Capability.Inject),
                [EsbCommands.Send,EsbCommands.Start, EsbCommands.Stop]
            ),
            Domain.LogitechUnifying : (
                (Capability.Inject),
                [UnifyingCommands.Send,UnifyingCommands.Start, UnifyingCommands.Stop]
            )
        }
    ),
    DLT_RESERVED_06 : (
        {
            Domain.Phy : (
                (Capability.Sniff | Capability.NoRawData),
                [PhyCommands.Sniff, PhyCommands.Start, PhyCommands.Stop, PhyCommands.SetFrequency, PhyCommands.GetSupportedFrequencies, PhyCommands.SetPacketSize, PhyCommands.SetDataRate, PhyCommands.SetEndianness, PhyCommands.SetSyncWord, PhyCommands.SetASKModulation, PhyCommands.SetFSKModulation, PhyCommands.SetGFSKModulation, PhyCommands.SetBPSKModulation, PhyCommands.SetQPSKModulation, PhyCommands.SetLoRaModulation]
            )
        },
        {
            Domain.Phy : (
                (Capability.Inject | Capability.NoRawData),
                [PhyCommands.Send,PhyCommands.Start, PhyCommands.Stop, PhyCommands.SetFrequency, PhyCommands.GetSupportedFrequencies, PhyCommands.SetPacketSize, PhyCommands.SetDataRate, PhyCommands.SetEndianness, PhyCommands.SetSyncWord, PhyCommands.SetASKModulation, PhyCommands.SetFSKModulation, PhyCommands.SetGFSKModulation, PhyCommands.SetBPSKModulation, PhyCommands.SetQPSKModulation, PhyCommands.SetLoRaModulation]
            )
        }
    )
}



import logging

from scapy.packet import Packet
from scapy.layers.zigbee import ZigbeeSecurityHeader
from whad.zigbee.connector import Zigbee
from whad.zigbee.sniffing import SnifferConfiguration, KeyExtractedEvent
from whad.zigbee.crypto import ZigbeeDecryptor, TouchlinkKeyManager, TransportKeyDistribution
from whad.exceptions import UnsupportedCapability
from whad.helpers import message_filter
from whad.common.sniffing import EventsManager
from whad.hub.dot15d4 import RawPduReceived, PduReceived
from whad.hub.message import AbstractPacket
from whad.exceptions import WhadDeviceDisconnected
from whad.device import WhadDevice

logger = logging.getLogger(__name__)

class Sniffer(Zigbee, EventsManager):
    """
    Zigbee Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device: WhadDevice):
        """Sniffer initialization.

        :param device: Device to use for sniffing
        :type device: WhadDevice
        """
        Zigbee.__init__(self, device)
        EventsManager.__init__(self)


        self.__configuration = SnifferConfiguration()
        self.__decryptor = ZigbeeDecryptor()
        self.__touchlink_key_derivation = TouchlinkKeyManager()
        self.__transport_key_distribution = TransportKeyDistribution()
        # Check if device can perform sniffing
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

    def _enable_sniffing(self):
        for key in self.__configuration.keys:
            self.__decryptor.add_key(key)
        self.sniff_zigbee(channel=self.__configuration.channel)

    def add_key(self, key: bytes):
        """Add an encryption key to our sniffer.

        :param key: encryption key to add
        :type key: bytes
        """
        self.__configuration.keys.append(key)

    def clear_keys(self):
        """Clear all stored encryption keys.
        """
        self.__configuration.keys = []

    @property
    def decrypt(self) -> bool:
        """Decryption enabled
        """
        return self.__configuration.decrypt

    @decrypt.setter
    def decrypt(self, decrypt: bool):
        """Set decryption status
        """
        self.__configuration.decrypt = decrypt


    @property
    def channel(self) -> int:
        """Current channel
        """
        return self.__configuration.channel

    @channel.setter
    def channel(self, channel: int = 11):
        """Set current channel.

        :param channel: new ZigBee channel to use
        :type channel: int
        """
        self.stop()
        self.__configuration.channel = channel
        self._enable_sniffing()


    @property
    def configuration(self) -> SnifferConfiguration:
        """Current sniffer configuration.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration):
        self.stop()
        self.__configuration = new_configuration
        self._enable_sniffing()

    def process_packet(self, packet: Packet):
        """Process received ZigBee packet.

        :param packet: received packet
        :type packet: :class:`scapy.packet.Packet`
        :return: received packet
        :rtype: :class:`scapy.packet.Packet`
        """
        if self.__touchlink_key_derivation.unencrypted_key is not None:
            logger.info("[i] New key extracted: %s",
                        self.__touchlink_key_derivation.unencrypted_key.hex())
            self.trigger_event(KeyExtractedEvent(self.__touchlink_key_derivation.unencrypted_key))
            self.add_key(self.__touchlink_key_derivation.unencrypted_key)
            self.__decryptor.add_key(self.__touchlink_key_derivation.unencrypted_key)
            self.__touchlink_key_derivation.reset()

        if self.__transport_key_distribution.transport_key is not None:
            logger.info(
                "[i] New key extracted: %s",
                self.__transport_key_distribution.transport_key.hex())
            self.trigger_event(KeyExtractedEvent(self.__transport_key_distribution.transport_key))
            self.add_key(self.__transport_key_distribution.transport_key)
            self.__decryptor.add_key(self.__transport_key_distribution.transport_key)
            self.__transport_key_distribution.reset()


        if ZigbeeSecurityHeader in packet and self.__configuration.decrypt:
            decrypted, success = self.__decryptor.attempt_to_decrypt(packet)
            if success:
                packet.data = decrypted
                packet.metadata.decrypted = True


        if self.__configuration.pairing:
            self.__touchlink_key_derivation.process_packet(packet)
            self.__transport_key_distribution.process_packet(packet)

        return packet

    def sniff(self):
        try:
            while True:
                if self.support_raw_pdu():
                    message_type = RawPduReceived
                else:
                    message_type = PduReceived

                message = self.wait_for_message(filter=message_filter(message_type), timeout=0.1)
                if message is not None and issubclass(message, AbstractPacket):
                    packet = message.to_packet()
                    self.monitor_packet_rx(packet)
                    packet = self.process_packet(packet)
                    yield packet
        except WhadDeviceDisconnected:
            return


"""WHAD Protocol Dot15d4 pdu messages abstraction layer.
"""
from struct import pack
from scapy.layers.dot15d4 import Dot15d4, Dot15d4FCS
from whad.hub.message import pb_bind, PbFieldInt, PbFieldBytes, PbMessageWrapper, \
    PbFieldBool
from whad.hub.dot15d4 import Dot15d4Domain, Dot15d4Metadata

@pb_bind(Dot15d4Domain, 'send', 1)
class SendPdu(PbMessageWrapper):
    """Send Dot15d4 PDU message class
    """
    channel = PbFieldInt('dot15d4.send.channel')
    pdu = PbFieldBytes('dot15d4.send.pdu')

    def to_packet(self):
        """Convert message to the corresponding scapy packet
        """
        return Dot15d4(self.pdu)

    @staticmethod
    def from_packet(packet, channel: int = 11):
        """Convert a scapy packet to a SendPdu message
        """
        if Dot15d4 in packet:
            pdu = bytes(packet[Dot15d4])
        elif Dot15d4FCS in packet:
            pdu = bytes(packet[Dot15d4FCS])[:-2]
        else:
            return None

        msg = SendPdu(
            channel=channel,
            pdu=pdu
        )
        return msg

        return msg

@pb_bind(Dot15d4Domain, 'send_raw', 1)
class SendRawPdu(PbMessageWrapper):
    """Send Dot15d4 raw PDU message class
    """
    channel = PbFieldInt('dot15d4.send_raw.channel')
    pdu = PbFieldBytes('dot15d4.send_raw.pdu')
    fcs = PbFieldInt('dot15d4.send_raw.fcs')

    def to_packet(self):
        """Convert message to the corresponding scapy packet
        """
        return Dot15d4FCS(self.pdu + bytes(pack('>H', self.fcs)))

    @staticmethod
    def from_packet(packet, channel: int = 11):
        """Convert a scapy packet to a SendPdu message
        """

        if Dot15d4FCS in packet:
            pdu = bytes(packet[Dot15d4FCS])[:-2]

        elif Dot15d4 in packet:
            packet.show()
            pdu = bytes(packet[Dot15d4])
        else:
            return None

        msg = SendRawPdu(
            channel=channel,
            pdu=pdu,
            fcs=packet.fcs
        )
        return msg

        return msg

@pb_bind(Dot15d4Domain, 'pdu', 1)
class PduReceived(PbMessageWrapper):
    """Dot15d4 PDU received message class
    """
    channel = PbFieldInt('dot15d4.pdu.channel')
    pdu = PbFieldBytes('dot15d4.pdu.pdu')
    rssi = PbFieldInt('dot15d4.pdu.rssi', optional=True)
    timestamp = PbFieldInt('dot15d4.pdu.timestamp', optional=True)
    fcs_validity = PbFieldBool('dot15d4.pdu.fcs_validity', optional=True)
    lqi = PbFieldInt('dot15d4.pdu.lqi', optional=True)

    def to_packet(self):
        """Convert message to its scapy packet representation
        """
        # Create packet
        packet = Dot15d4(bytes(self.pdu))

        # Set packet metadata
        packet.metadata = Dot15d4Metadata()
        packet.metadata.channel = self.channel

        packet.metadata.decrypted = False

        if self.lqi is not None:
            packet.metadata.lqi = self.lqi
        if self.rssi is not None:
            packet.metadata.rssi = self.rssi
        if self.timestamp is not None:
            packet.metadata.timestamp = self.timestamp
        if self.fcs_validity is not None:
            packet.metadata.is_fcs_valid = self.fcs_validity

        # Return packet
        return packet


    @staticmethod
    def from_packet(packet):
        """Convert scapy packet to a PduReceived message
        """
        # Create a PduReceived message
        msg = PduReceived(
            channel=packet.metadata.channel,
            pdu=bytes(packet.getlayer(Dot15d4)),
        )
        # Add optional metadata

        if packet.metadata.decrypted is not None:
            msg.decrypted = packet.metadata.decrypted
        if packet.metadata.lqi is not None:
            msg.lqi = packet.metadata.lqi
        if packet.metadata.rssi is not None:
            msg.rssi = packet.metadata.rssi
        if packet.metadata.timestamp is not None:
            msg.rssi = packet.metadata.timestamp
        if packet.metadata.is_fcs_valid is not None:
            msg.fcs_validity = packet.metadata.is_fcs_valid

        # Return metadata
        return msg


@pb_bind(Dot15d4Domain, 'raw_pdu', 1)
class RawPduReceived(PbMessageWrapper):
    """Dot15d4 raw PDU received message class
    """
    channel = PbFieldInt('dot15d4.raw_pdu.channel')
    pdu = PbFieldBytes('dot15d4.raw_pdu.pdu')
    fcs = PbFieldInt('dot15d4.raw_pdu.fcs')
    rssi = PbFieldInt('dot15d4.raw_pdu.rssi', optional=True)
    timestamp = PbFieldInt('dot15d4.raw_pdu.timestamp', optional=True)
    fcs_validity = PbFieldBool('dot15d4.raw_pdu.fcs_validity', optional=True)
    lqi = PbFieldInt('dot15d4.raw_pdu.lqi', optional=True)


    def to_packet(self):
        """Convert message to scapy packet
        """

        # Create packet
        #print('converting %s' % (self.pdu + bytes(pack(">H", self.fcs))).hex())
        packet = Dot15d4FCS(bytes(self.pdu) + bytes(pack(">H", self.fcs)))

        # Set packet metadata
        packet.metadata = Dot15d4Metadata()
        packet.metadata.channel = self.channel
        packet.metadata.decrypted = False
        if self.lqi is not None:
            packet.metadata.lqi = self.lqi
        if self.rssi is not None:
            packet.metadata.rssi = self.rssi
        if self.timestamp is not None:
            packet.metadata.timestamp = self.timestamp
        if self.fcs_validity is not None:
            packet.metadata.is_fcs_valid = self.fcs_validity

        # Return packet
        return packet

    @staticmethod
    def from_packet(packet):
        """Convert packet to a RawPduReceived message.
        """
        # Create a PduReceived message
        msg = RawPduReceived(
            channel=packet.metadata.channel,
            pdu=bytes(packet.getlayer(Dot15d4FCS))[:-2],
            fcs=packet.fcs
        )

        # Add optional metadata

        if packet.metadata.decrypted is not None:
            msg.decrypted = packet.metadata.decrypted
        if packet.metadata.lqi is not None:
            msg.lqi = packet.metadata.lqi
        if packet.metadata.rssi is not None:
            msg.rssi = packet.metadata.rssi
        if packet.metadata.timestamp is not None:
            msg.timestamp = packet.metadata.timestamp
        if packet.metadata.is_fcs_valid is not None:
            msg.fcs_validity = packet.metadata.is_fcs_valid

        # Return metadata
        return msg
